# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import logging
import argparse
import sys

from plumbery.engine import PlumberyEngine
from plumbery import __version__
from plumbery.plogging import plogging


def parse_args(args=[]):
    """
    Guesses the intention of the runner of this program

    :param args: arguments to be considered for this invocation
    :type args: a list of ``str``

    You have to run the following command to know more::

        $ python -m plumbery fittings.yaml -h

    """

    parser = argparse.ArgumentParser(
        prog='python -m plumbery',
        description='Plumbing infrastructure with Apache Libcloud.',
        epilog='example: python -m plumbery fittings.yaml build')

    parser.add_argument(
        'fittings',
        nargs=1,
        help="File that is containing fittings plan, or '-' to read stdin")

    parser.add_argument(
        'action',
        nargs=1,
        help="An action, or a polisher: 'deploy', 'refresh', dispose', "
             "'secrets', 'build', 'configure', 'start', 'prepare', "
             "'information', 'ping', 'inventory', 'ansible', "
             "'stop', 'wipe', 'destroy'")

    parser.add_argument(
        'tokens',
        nargs='*',
        help="One blueprint, or several, e.g., 'web' or 'web sql'."
             "If omitted, all blueprints will be considered. "
             "Zero or more locations, e.g., '@NA12'. "
             "If omitted, all locations will be considered.",
        default=None)

    parser.add_argument(
        '-p', '--parameters', nargs='*',
        help='Parameters for this fittings plan')

    parser.add_argument(
        '-s', '--safe',
        help='Safe mode, no actual change is made to the infrastructure',
        action='store_true')

    group = parser.add_mutually_exclusive_group()

    group.add_argument(
        '-d', '--debug',
        help='Log as much information as possible',
        action='store_true')

    group.add_argument(
        '-q', '--quiet',
        help='Silent mode, log only warnings and errors',
        action='store_true')

    parser.add_argument(
        '-v', '--version',
        help='Print version of this software',
        action='version',
        version='plumbery ' + __version__)

    args = parser.parse_args(args)

    if args.debug:
        plogging.setLevel(logging.DEBUG)
    elif args.quiet:
        plogging.setLevel(logging.WARNING)
    else:
        plogging.setLevel(logging.INFO)

    if 'version' in args:
        print(args.version)

    args.fittings = args.fittings[0]
    plogging.debug("- loading '{}'".format(args.fittings))

    args.action = args.action[0].lower()

    args.blueprints = []
    args.facilities = []
    for token in args.tokens:
        if token[0] == '@':
            if token == '@':
                raise ValueError("Missing location after @. "
                                 "Correct example: '@AU11'")
            args.facilities.append(token[1:])
        else:
            args.blueprints.append(token)

    if len(args.blueprints) < 1:
        args.blueprints = None
    else:
        plogging.debug('blueprints: '+' '.join(args.blueprints))

    if len(args.facilities) < 1:
        args.facilities = None
    else:
        plogging.debug('facilities: '+' '.join(args.facilities))

    return args


def main(args=None, engine=None):
    """
    Runs plumbery from the command line

    :param args: arguments to be considered for this invocation
    :type args: a list of ``str``

    :param engine: an instance of the plumbery engine
    :type engine: :class:`plumbery.PlumberEngine`

    Example::

        $ python -m plumbery fittings.yaml build web

    In this example, plumbery loads fittings plan from ``fittings.yaml``, then
    it builds the blueprint named ``web``.

    If no blueprint is mentioned, then plumbery looks at all blueprint
    definitions in the fittings plan. In other terms, the following command
    builds the entire fittings plan, eventually across multiple facilities::

        $ python -m plumbery fittings.yaml build

    Of course, plumbery can be invoked through the entire life cycle of your
    fittings::

        $ python -m plumbery fittings.yaml build
        $ python -m plumbery fittings.yaml start
        $ python -m plumbery fittings.yaml polish

        ... nodes are up and running ...

        $ python -m plumbery fittings.yaml stop

        ... nodes have been stopped ...

        $ python -m plumbery fittings.yaml wipe

        ... nodes have been destroyed, but the infrastructure remains ...

        $ python -m plumbery fittings.yaml destroy

        ... every virtual resources has been removed ...


    To focus at a single location, put the character '@' followed by the id.
    For example, to build fittings only at 'NA12' you would type::

        $ python -m plumbery fittings.yaml build @NA12

    To focus on one blueprint just mention its name on the command line.
    For example, if fittings plan has a blueprint for nodes running Docker,
    then you may use following statements to bootstrap each node::

        $ python -m plumbery fittings.yaml build docker
        $ python -m plumbery fittings.yaml start docker
        $ python -m plumbery fittings.yaml prepare docker

        ... Docker is up and running at multiple nodes ...

    If you create a new polisher and put it in the directory
    ``plumbery\polishers``, then it will become automatically available::

        $ python -m plumbery fittings.yaml my_special_stuff

    To get some help, you can type::

        $ python -m plumbery -h

    """

    # part 1 - understand what the user wants

    if args is None:
        args = sys.argv[1:]

    try:
        args = parse_args(args)

    except Exception as feedback:
        plogging.error("Incorrect arguments. "
                      "Maybe the following can help: python -m plumbery -h")
        if plogging.getEffectiveLevel() == logging.DEBUG:
            raise
        else:
            plogging.error("{}: {}".format(
                feedback.__class__.__name__,
                str(feedback)))
        sys.exit(2)

    # part 2 - get a valid and configured engine

    if engine is None:
        try:
            engine = PlumberyEngine(args.fittings, args.parameters)

            if args.safe:
                engine.safeMode = True

        except Exception as feedback:
            if plogging.getEffectiveLevel() == logging.DEBUG:
                plogging.error("Cannot read fittings plan from '{}'".format(
                    args.fittings))
                raise
            else:
                plogging.error("Cannot read fittings plan from '{}'"
                              ", run with -d for debug".format(
                                  args.fittings))
                plogging.error("{}: {}".format(
                    feedback.__class__.__name__,
                    str(feedback)))
            sys.exit(2)

    # part 3 - do the job

    try:
        engine.do(args.action, args.blueprints, args.facilities)

        plogging.info(engine.document_elapsed())

    except Exception as feedback:
        if plogging.getEffectiveLevel() == logging.DEBUG:
            plogging.error("Unable to do '{}'".format(args.action))
            raise
        else:
            plogging.error("Unable to do '{}', run with -d for debug".format(
                args.action))
            plogging.error("{}: {}".format(
                feedback.__class__.__name__,
                str(feedback)))
        sys.exit(1)

if __name__ == "__main__":
    try:
        main()

        # if some errors have been logged, make it explicit to the caller
        if plogging.foundErrors():
            plogging.error("Hit some error, you should check the logs")
            sys.exit(1)

    except KeyboardInterrupt:
        plogging.error("Aborted by user")
        sys.exit(1)



#!/usr/bin/env python

"""
Tests for `terraform` module.
"""

import unittest
import os
import mock

from plumbery.terraform import Terraform

from tempfile import gettempdir
tempdir = gettempdir()


class TerraformTests(unittest.TestCase):
    @mock.patch('os.getenv', mock.Mock(return_value=None))
    def test_init_no_env(self):
        with self.assertRaises(RuntimeError):
            t = Terraform(os.getcwd())
            t.graph(os.getcwd())

    def test_build(self):
        t = Terraform(tempdir)
        t._run_tf = mock.MagicMock(return_value=(2, "", ""))
        t.build({})
        t._run_tf.assert_called_with("apply", os.path.join(tempdir, '.tfstate'))

    def test_destroy(self):
        t = Terraform(tempdir)
        t._run_tf = mock.MagicMock(return_value=(2, "", ""))
        t.destroy({})
        t._run_tf.assert_called_with("plan", tempdir, destroy=True, detailed_exitcode=True, input=False,
                                     var_file=os.path.join(tempdir, '.tfvars'))

    def test_graph(self):
        t = Terraform(tempdir)
        t._run_tf = mock.MagicMock(return_value=(2, "", ""))
        t.graph(tempdir)
        t._run_tf.assert_called_with("graph", tempdir)

if __name__ == '__main__':
    import sys
    sys.exit(unittest.main())


# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys

from libcloud.compute.base import NodeState

from plumbery.polisher import PlumberyPolisher
from plumbery.text import PlumberyText
from plumbery.text import PlumberyContext
from plumbery.text import PlumberyNodeContext
from plumbery.plogging import plogging


class InformationPolisher(PlumberyPolisher):
    """
    Shows information attached to fittings plan, to containers, to nodes

    """

    def go(self, engine):
        """
        Lists information registered at the highest level of fittings plan

        :param engine: the plumbery engine itself
        :type engine: :class:`plumbery.PlumberyEngine`

        """

        self.engine = engine

        self.information = []

        environment = PlumberyContext(context=self.engine)

        lines = []
        for line in engine.information:

            tokens = line.split(' ')
            if tokens[0] == 'echo':
                tokens.pop(0)
            message = ' '.join(tokens)
            message = PlumberyText.expand_string(
                message, environment)
            lines.append(message)

        if len(lines) < 1:
            return

        self.information.append("About this fittings plan:")

        for line in lines:
            self.information.append("- {}".format(line))

    def move_to(self, facility):
        """
        Lists information registered in one document of fittings plan

        :param facility: a target facility
        :type facility: :class:`plumbery.PlumberyFacility`

        This function is called once for each facility that is visited during
        the polishing process. You can override it for any specific
        initialisation that you would require.

        """

        self.facility = facility

        environment = PlumberyContext(context=self.facility)

        if ('information' in facility.settings and
                isinstance(facility.settings['information'], str)):

            facility.settings['information'] = \
                facility.settings['information'].strip('\n').split('\n')

        lines = []
        if ('information' in facility.settings and
                isinstance(facility.settings['information'], list) and
                len(facility.settings['information']) > 0):

            for line in facility.settings['information']:

                tokens = line.split(' ')
                if tokens[0] == 'echo':
                    tokens.pop(0)
                message = ' '.join(tokens)
                message = PlumberyText.expand_string(
                    message, environment)
                lines.append(message)

        if len(lines) < 1:
            return

        self.information.append("About '{}':".format(
            facility.get_location_id()))

        for line in lines:
            self.information.append("- {}".format(line))

    def shine_container(self, container):
        """
        Lists information registered at the container level

        :param container: the container to be polished
        :type container: :class:`plumbery.PlumberyInfrastructure`

        """

        environment = PlumberyNodeContext(node=None,
                                          container=container,
                                          context=self.facility)

        if ('information' in container.blueprint.keys() and
                isinstance(container.blueprint['information'], str)):

            container.blueprint['information'] = \
                container.blueprint['information'].strip('\n').split('\n')

        lines = []
        if ('information' in container.blueprint.keys() and
                isinstance(container.blueprint['information'], list) and
                len(container.blueprint['information']) > 0):

            for line in container.blueprint['information']:

                tokens = line.split(' ')
                if tokens[0] == 'echo':
                    tokens.pop(0)
                message = ' '.join(tokens)
                message = PlumberyText.expand_string(
                    message, environment)
                lines.append(message)

        if len(lines) > 0:

            self.information.append("About '{}':".format(
                container.blueprint['target']))

            for line in lines:
                self.information.append("- {}".format(line))

        if 'balancers' in container.blueprint.keys():

            for item in container.blueprint['balancers']:

                if isinstance(item, dict):
                    label = list(item)[0]
                    settings = item[label]
                else:
                    label = str(item)
                    settings = {}

                if 'information' not in settings:
                    continue

                name = container.name_balancer(label, settings)
                balancer = container._get_balancer(name)
                print(balancer)

                context = {
                    'balancer.name': name,
                    'balancer.ip': balancer.ip,
                    'balancer.port': balancer.port,
                    }

                environment = PlumberyContext(dictionary=context,
                                              context=self.facility)

                lines = []
                for line in settings['information']:

                    message = PlumberyText.expand_string(
                        line, environment)
                    lines.append(message)

                if len(lines) > 0:

                    self.information.append("About '{}':".format(name))

                    for line in lines:
                        self.information.append("- {}".format(line))

    def list_information(self, node, settings, container):
        """
        Lists information for a node if it exists

        :param node: the node to be polished
        :type node: :class:`libcloud.compute.base.Node`

        :param settings: the fittings plan for this node
        :type settings: ``dict``

        :param container: the container of this node
        :type container: :class:`plumbery.PlumberyInfrastructure`

        """

        environment = PlumberyNodeContext(node=node,
                                          container=container,
                                          context=self.facility)

        if ('information' in settings and
                isinstance(settings['information'], str)):

            settings['information'] = \
                settings['information'].strip('\n').split('\n')

        information = []
        if ('information' in settings
                and isinstance(settings['information'], list)
                and len(settings['information']) > 0):

            for line in settings['information']:

                tokens = line.split(' ')
                if tokens[0] == 'echo':
                    tokens.pop(0)
                message = ' '.join(tokens)
                message = PlumberyText.expand_string(
                    message, environment)
                information.append(message)

        return information

    def shine_node(self, node, settings, container):
        """
        Shows information attached to node, if any

        :param node: the node to be polished
        :type node: :class:`libcloud.compute.base.Node`

        :param settings: the fittings plan for this node
        :type settings: ``dict``

        :param container: the container of this node
        :type container: :class:`plumbery.PlumberyInfrastructure`

        """

        plogging.info("- examinating node '{}'".format(settings['name']))

        lines = []

        if (node is not None and 'description' in node.extra):
            description = node.extra['description'].replace(
                '#plumbery', '').strip()
            if len(description) > 0:
                lines.append(description)

        if node is None:
            plogging.debug("- not found")
            lines.append("node is unknown")
        elif node.state == NodeState.RUNNING:
            lines.append("node is up and running")
        elif node.state in [NodeState.TERMINATED,
                            NodeState.STOPPED,
                            NodeState.SUSPENDED]:
            lines.append("node has been stopped")
        else:
            lines.append("state: {}".format(node.state))

        if node is not None:
            lines += self.list_information(
                node=node, settings=settings, container=container)

        if len(lines) < 1:
            return

        self.information.append("About '{}':".format(settings['name']))

        for line in lines:
            self.information.append("- {}".format(line))

    def reap(self):
        """
        Saves information gathered through the polishing sequence

        All information captured in dumped in a file, in YAML format,
        to provide a flexible and accurate inventory of all live nodes
        described in the fittings plan.

        """

        if len(self.information) < 1:
            return

        if 'output' in self.settings:
            fileName = self.settings['output']
            plogging.info("Writing information in '{}'".format(fileName))
            stream = open(fileName, 'w')
            stream.write('\n'.join(self.information)+'\n')
        else:
            plogging.info('\n'.join(self.information)+'\n')




# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from plumbery.plogging import plogging

__all__ = ['PlumberyFitting', 'PlumberyFittingLoader']


class PlumberyFitting(object):
    """
    Represents some fitting

    :param engine: the automate that is coordinating
            plumbing activities at multiple facilities
    :type engine: :class:`plumbery.PlumberyEngine`

    :param facility: the place that is making a context for this fitting
    :type facility: :class:`plumbery.PlumberyFacility`

    Fittings are representation of cloud services, such as 'domain' or
    'node'. They are provided with settings that should be parsed and checked.
    A `ValueError` should be raised in case of invalid settings.

    """

    def __init__(self, engine, facility):
        self.engine = engine
        self.facility = facility

    def parse(self, settings):
        """
        Parses and checks settings

        :param settings: specific settings for this fittings
        :type settings: ``dict``

        This function raises a `ValueError` if invalid settings are provided.
        """

        raise NotImplementedError()

    def do(self, action, parameters=None):
        """
        Performs some action

        :param action: the action to perform
        :type action: ``str``

        :param parameters: the additional parameters, if any
        :type parameters: ``dict``

        This function dispatches the action to the member function of the
        same name. For example, ``fittings.do('start')`` will execute
        ``fittings.do_start()``.
        """

        method = getattr(self, 'do_'+action, None)
        if callable(method):
            return method(parameters)
        return True


class PlumberyFittingLoader(object):
    """
    Loads and set a fitting classe dynamically

    You can create fittings of your own, or use fittings from other
    persons. All fittings have to be placed in the directory
    ``plumbery/fittings``. Each fitting should be put in a separate
    python file, and define a class that repeats the file name. For example
    the file::

        plumbery/fittings/domain.py

    should contain::

        class DomainFitting(PlumberyFitting):
        ...

    Once this is done properly, plumbery will be able to parse related
    directives, and to process actions accordingly.

    """

    @classmethod
    def from_shelf(cls, label, engine, facility, settings={}):
        """
        Picks up fitting from the shelf

        :param label: name of fitting to use, e.g., ``domain``
        :type label: ``str``

        :param settings: specific settings for this fitting
        :type param: ``dict``

        :param engine: the automate that is coordinating
                plumbing activities at multiple facilities
        :type engine: :class:`plumbery.PlumberyEngine`

        :param facility: the place that is making a context for this fitting
        :type facility: :class:`plumbery.PlumberyFacility`

        :return: instance of fitting ready to use
        :rtype: :class:`plumbery.PlumberyFitting`

        """

        moduleName = 'plumbery.fittings.' + label
        fittingsName = label.capitalize() + 'Fitting'
        try:

            plogging.debug("Importing '{}'".format(moduleName))
            fittingsModule = __import__(
                moduleName,
                globals(),
                locals(),
                [fittingsName])

            plogging.debug("Instantiating '{}'".format(fittingsName))
            fittingsClass = getattr(fittingsModule, fittingsName)
            if settings is None:
                settings = {}
            fitting = fittingsClass(engine, facility)
            fitting.parse(settings)
            fitting.label = label
            return fitting

        except ImportError as feedback:
            plogging.debug("Unable to find module '{}'".format(moduleName))
            raise

        except TypeError as feedback:
            plogging.debug("Invalid settings for '{}'".format(moduleName))
            raise

        except ValueError as feedback:
            plogging.debug("Invalid settings for '{}'".format(moduleName))
            raise

        except Exception as feedback:
            plogging.debug("Unable to import '{}' from '{}'".format(
                fittingsName, moduleName))
            raise



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

from plumbery.polisher import PlumberyPolisher
from plumbery.nodes import PlumberyNodes
from plumbery.exception import ConfigurationError
from plumbery.polishers.memory import MemoryConfiguration
from plumbery.polishers.cpu import CpuConfiguration
from plumbery.polishers.monitoring import MonitoringConfiguration
from plumbery.polishers.disks import DisksConfiguration
from plumbery.polishers.backup import BackupConfiguration
from plumbery.polishers.windows import WindowsConfiguration
from plumbery.plogging import plogging


class ConfigurePolisher(PlumberyPolisher):
    """
    Configures various elements in fittings plan

    This polisher looks at each node in sequence, and adjust settings
    according to fittings plan. This is covering various features that
    can not be set during the creation of nodes, such as:
    - number of CPU
    - quantity of RAM
    - monitoring
    - network interfaces

    """

    configuration_props = (MonitoringConfiguration,
                           DisksConfiguration, BackupConfiguration,
                           WindowsConfiguration)

    def move_to(self, facility):
        """
        Moves to another API endpoint

        :param facility: access to local parameters and functions
        :type facility: :class:`plumbery.PlumberyFacility`


        """

        self.facility = facility
        self.region = facility.region
        self.nodes = PlumberyNodes(facility)

    def shine_container(self, container):
        """
        Configures a container

        :param container: the container to be polished
        :type container: :class:`plumbery.PlumberyInfrastructure`

        """

        plogging.info("Configuring blueprint '{}'".format(
            container.blueprint['target']))

        if container.network is None:
            plogging.error("- aborted - no network here")
            return

        self.container = container

        plogging.info("- waiting for nodes to be deployed")

        names = self.nodes.list_nodes(container.blueprint)
        for name in sorted(names):
            while True:
                node = self.nodes.get_node(name)
                if node is None:
                    plogging.error("- aborted - missing node '{}'".format(name))
                    return

                if node.extra['status'].action is None:
                    plogging.debug("- {} is ready".format(node.name))
                    break

                if (node is not None
                        and node.extra['status'].failure_reason is not None):

                    plogging.error("- aborted - failed deployment "
                                 "of node '{}'".format(name))
                    return

                time.sleep(20)

        plogging.info("- nodes have been deployed")

        container._build_firewall_rules()

        container._build_balancer()

    def set_node_compute(self, node, cpu, memory):
        """
        Sets compute capability

        :param node: the node to be polished
        :type node: :class:`libcloud.compute.base.Node`

        :param cpu: the cpu specification
        :type cpu: ``DimensionDataServerCpuSpecification``

        :param memory: the memory size, expressed in Giga bytes
        :type memory: ``int``

        """

        changed = False

        if cpu is not None and 'cpu' in node.extra:

            if int(cpu.cpu_count) != int(node.extra['cpu'].cpu_count):
                plogging.info("- changing to {} cpu".format(
                    cpu.cpu_count))
                changed = True

            if (int(cpu.cores_per_socket) !=
                    int(node.extra['cpu'].cores_per_socket)):

                plogging.info("- changing to {} core(s) per socket".format(
                    cpu.cores_per_socket))
                changed = True

            if cpu.performance != node.extra['cpu'].performance:
                plogging.info("- changing to '{}' cpu performance".format(
                    cpu.performance.lower()))
                changed = True

        if memory is not None and 'memoryMb' in node.extra:

            if memory != int(node.extra['memoryMb']/1024):
                plogging.info("- changing to {} GB memory".format(
                    memory))
                changed = True

        if not changed:
            plogging.debug("- no change in compute")
            return

        if self.engine.safeMode:
            plogging.info("- skipped - safe mode")
            return

        while True:
            try:
                self.region.ex_reconfigure_node(
                    node=node,
                    memory_gb=memory,
                    cpu_count=cpu.cpu_count,
                    cores_per_socket=cpu.cores_per_socket,
                    cpu_performance=cpu.performance)

                plogging.info("- in progress")

            except Exception as feedback:
                if 'RESOURCE_BUSY' in str(feedback):
                    time.sleep(10)
                    continue

                if 'Please try again later' in str(feedback):
                    time.sleep(10)
                    continue

                plogging.info("- unable to reconfigure node")
                plogging.error(str(feedback))

            break

    def attach_node(self, node, networks):
        """
        Glues a node to multiple networks

        :param node: the target node
        :type node: :class:`libcloud.compute.base.Node`

        :param networks: a list of networks to connect, and ``internet``
        :type networks: list of ``str``

        This function adds network interfaces to a node, or adds address
        translation to the public Internet.

        Example in the fittings plan::

          - web:
              domain:
                ipv4: 6
              ethernet:
                name: gigafox.data
              nodes:
                - web[10..12]:
                    glue:
                      - gigafox.control
                      - internet 80 443

        In this example, another network interface is added to each node for
        connection to the Ethernet network ``gigafox.control``.

        Also, public IPv4 addresses are mapped on private addresses, so that
        each node web10, web11 and web12 is reachable from the internet.
        Public IPv4 addresses are taken from pool declared at the domain level,
        with the attribute ``ipv4``. In the example above, 6 addresses are
        assigned to the network domain, of which 3 are given to web nodes.

        If one or multiple numbers are mentioned after the keyword `internet`,
        they are used to configure the firewall appropriately.

        """

        hasChanged = False

        if node is None:
            return hasChanged

        for line in networks:

            tokens = line.strip(' ').split(' ')
            token = tokens.pop(0)

            if token.lower() == 'internet':
                self.attach_node_to_internet(node, tokens)
                continue

            if token == self.container.blueprint['ethernet']['name']:
                continue

            if token.lower() == 'primary':
                continue

            plogging.info("Glueing node '{}' to network '{}'"
                         .format(node.name, token))
            vlan = self.container.get_ethernet(token.split('::'))
            if vlan is None:
                plogging.info("- network '{}' is unknown".format(token))
                continue

            kwargs = {}
            if len(tokens) > 0:

                numbers = tokens.pop(0).strip('.').split('.')
                subnet = vlan.private_ipv4_range_address.split('.')
                while len(numbers) < 4:
                    numbers.insert(0, subnet[3-len(numbers)])

                private_ipv4 = '.'.join(numbers)
                plogging.debug("- using address '{}'".format(private_ipv4))
                kwargs['private_ipv4'] = private_ipv4

            if self.engine.safeMode:
                plogging.info("- skipped - safe mode")
                continue

            if 'private_ipv4' not in kwargs:
                kwargs['vlan'] = vlan

            while True:
                try:
                    self.region.ex_attach_node_to_vlan(node, **kwargs)
                    plogging.info("- in progress")
                    hasChanged = True

                except Exception as feedback:

                    if 'RESOURCE_BUSY' in str(feedback):
                        time.sleep(10)
                        continue

                    elif 'RESOURCE_LOCKED' in str(feedback):
                        plogging.info("- not now - locked")

                    elif 'INVALID_INPUT_DATA' in str(feedback):
                        plogging.info("- already there")

                    else:
                        plogging.info("- unable to glue node")
                        plogging.error(str(feedback))

                break

        return hasChanged

    def attach_node_to_internet(self, node, ports=[]):
        """
        Adds address translation for one node

        :param node: node that has to be reachable from the internet
        :type node: :class:`libcloud.common.Node`

        :param ports: the ports that have to be opened
        :type ports: a list of ``str``

        """

        plogging.info("Making node '{}' reachable from the internet"
                     .format(node.name))

        domain = self.container.get_network_domain(
            self.container.blueprint['domain']['name'])

        internal_ip = node.private_ips[0]

        external_ip = None
        for rule in self.region.ex_list_nat_rules(domain):
            if rule.internal_ip == internal_ip:
                external_ip = rule.external_ip
                plogging.info("- node is reachable at '{}'".format(external_ip))

        if self.engine.safeMode:
            plogging.info("- skipped - safe mode")
            return

        if external_ip is None:
            external_ip = self.container._get_ipv4()

            if external_ip is None:
                plogging.info("- no more ipv4 address available -- assign more")
                return

            while True:
                try:
                    self.region.ex_create_nat_rule(
                        domain,
                        internal_ip,
                        external_ip)
                    plogging.info("- node is reachable at '{}'".format(
                        external_ip))

                except Exception as feedback:
                    if 'RESOURCE_BUSY' in str(feedback):
                        time.sleep(10)
                        continue

                    elif 'RESOURCE_LOCKED' in str(feedback):
                        plogging.info("- not now - locked")
                        return

                    else:
                        plogging.info("- unable to add address translation")
                        plogging.error(str(feedback))

                break

        candidates = self.container._list_candidate_firewall_rules(node, ports)

        for rule in self.container._list_firewall_rules():

            if rule.name in candidates.keys():
                plogging.info("Creating firewall rule '{}'"
                             .format(rule.name))
                plogging.info("- already there")
                candidates = {k: candidates[k]
                              for k in candidates if k != rule.name}

        for name, rule in candidates.items():

            plogging.info("Creating firewall rule '{}'"
                         .format(name))

            if self.engine.safeMode:
                plogging.info("- skipped - safe mode")

            else:

                try:

                    self.container._ex_create_firewall_rule(
                        network_domain=domain,
                        rule=rule,
                        position='LAST')

                    plogging.info("- in progress")

                except Exception as feedback:

                    if 'NAME_NOT_UNIQUE' in str(feedback):
                        plogging.info("- already there")

                    else:
                        plogging.info("- unable to create firewall rule")
                        plogging.error(str(feedback))

    def shine_node(self, node, settings, container):
        """
        Finalizes setup of one node

        :param node: the node to be polished
        :type node: :class:`libcloud.compute.base.Node`

        :param settings: the fittings plan for this node
        :type settings: ``dict``

        :param container: the container of this node
        :type container: :class:`plumbery.PlumberyInfrastructure`

        """

        plogging.info("Configuring node '{}'".format(settings['name']))
        if node is None:
            plogging.info("- not found")
            return

        try:
            cpu_prop = CpuConfiguration()
            cpu_prop.validate(settings)
            cpu = cpu_prop.configure(node, settings)

            ram_prop = MemoryConfiguration()
            ram_prop.validate(settings)
            memory = ram_prop.configure(node, settings)

            if memory is not False and cpu is not False:
                self.set_node_compute(node, cpu, memory)

        except ConfigurationError as ce:

            if self.engine.safeMode:
                plogging.warn(ce.message)
            else:
                raise ce

        for prop_cls in self.configuration_props:

            try:
                configuration_prop = prop_cls(engine=container.facility.plumbery,
                                              facility=self.facility)
                configuration_prop.validate(settings)
                configuration_prop.configure(node, settings)

            except ConfigurationError as ce:
                if self.engine.safeMode:
                    plogging.warn(ce.message)
                else:
                    raise ce

        container._add_to_pool(node)

        if 'glue' in settings:
            self.attach_node(node, settings['glue'])



"""
Tests for `facility` module.
"""

# special construct to allow relative import
#
if __name__ == "__main__" and __package__ is None:
    __package__ = "tests"
from tests import dummy

import unittest

from libcloud.compute.drivers.dimensiondata import DimensionDataNodeDriver

from plumbery.action import PlumberyAction
from plumbery.engine import PlumberyEngine
from plumbery.facility import PlumberyFacility

from .mock_api import DimensionDataMockHttp
DIMENSIONDATA_PARAMS = ('user', 'password')


class FakeElement:

    def find(self, dummy):
        return {'ipv6': 'nuts'}

# should be removed - head


class FakeConnection:

    object = FakeElement()

    def request_with_orgId_api_2(self, dummy):
        return self


class FakeDomain:

    id = 123


class FakeImage:

    name = 'RedHat 6 64-bit 4 CPU'


class FakeLocation:

    id = 'EU7'
    name = 'data centre in Amsterdam'
    country = 'Netherlands'


class FakeNetwork:

    id = 123


class FakeNode:

    id = 123
    name = 'stackstorm'
    extra = {'datacenterId': 'EU7'}


class FakeRegion:

    connection = FakeConnection()

    def create_node(self, name, image, auth, ex_network_domain, ex_vlan,
                    ex_is_started, ex_description):
        return True

    def ex_create_network_domain(self, location, name, service_plan,
                                 description):
        return FakeDomain()

    def ex_create_vlan(self, network_domain, name, private_ipv4_base_address,
                       description):
        return FakeNetwork()

    def ex_get_location_by_id(self, location):
        return FakeLocation()

    def ex_get_network_domain(self, location, network_domain):
        return []

    def ex_get_vlan(self, vlan_id):
        return FakeNetwork()

    def ex_list_network_domains(self, location):
        return []

    def ex_list_vlans(self, location):
        return []

    def ex_start_node(self, node):
        return True

    def ex_shutdown_graceful(self, node):
        return True

    def ex_wait_for_state(self, state, func, poll_interval=2, timeout=60,
                          *args, **kwargs):
        return []

    def destroy_node(self, node):
        return True

    def list_images(self, location):
        return [FakeImage()]

    def list_nodes(self):
        return [FakeNode()]


# should be removed - end

fakeFittings = {
    'regionId': 'dd-na',
    'locationId': 'NA9',
    'basement': 'fake1 unknown',
    'blueprints': [{
        'fake1': {
            'domain': {
                'name': 'VDC1',
                'service': 'ADVANCED',
                'description': 'fake'},
            'ethernet': {
                'name': 'vlan1',
                'subnet': '10.0.10.0',
                'description': 'fake'},
            'nodes': [{
                'stackstorm1': {
                    'description': 'fake',
                    'appliance': 'RedHat 6 64-bit 4 CPU'
                    }
                }]
            }
        }, {
        'fake2': {
            'domain': {
                'name': 'VDC1',
                'service': 'ADVANCED',
                'description': 'fake'},
            'ethernet': {
                'name': 'vlan1',
                'subnet': '10.0.10.0',
                'description': 'fake'},
            'nodes': [{
                'stackstorm2': {
                    'description': 'fake',
                    'appliance': 'RedHat 6 64-bit 4 CPU'
                    }
                }]
            }
        }, {
        'macro': 'unknown fake2'
        }]
    }

defaultsPlan = """
---
safeMode: True

defaults:
  locationId: EU6
  regionId: dd-eu
  domain:
    name: fake
    ipv4: auto
  ethernet:
    name: myVLAN
    subnet: 10.1.10.0
  bee:
    information:
      - one line of information
    cpu: 3
    memory: 6
    cloud-config:
      disable_root: true
      packages:
        - ntp
      runcmd:
        - echo "hello"
  cloud-config:
    disable_root: false
    ssh_pwauth: true
    packages:
      - foo
      - bar
    runcmd:
      - echo "romeo"
      - echo "juliet"
    ssh_keys:
      rsa_private: |
        {{ pair1.rsa_private }}

      rsa_public: "{{ pair1.ssh.rsa_public }}"

---

basement: myBlueprint

blueprints:

  - myBlueprint:
      domain:
        name: myDC
      ethernet:
        accept:
          - NA19::remoteNetwork
      nodes:
        - myServer:
            default: bee
            information:
              - complementary information
            memory: 5
            cloud-config:
              packages:
                - smtp
              runcmd:
                - echo "world"
"""


class FakeAction(PlumberyAction):
    def __init__(self, settings):
        self.count = 3

    def begin(self, engine):
        self.count += 100

    def enter(self, facility):
        self.count *= 2

    def process(self, blueprint):
        self.count += 5

    def quit(self):
        self.count -= 2

    def end(self):
        self.count += 1

class TestPlumberyFacility(unittest.TestCase):

    def setUp(self):
        self.plumbery = PlumberyEngine()
        self.plumbery.set_user_name('fake_user')
        self.plumbery.set_user_password('fake_password')
        DimensionDataNodeDriver.connectionCls.conn_classes = (
            None, DimensionDataMockHttp)
        DimensionDataMockHttp.type = None
        self.plumbery.region = DimensionDataNodeDriver(*DIMENSIONDATA_PARAMS)
        self.facility = PlumberyFacility(
            plumbery=self.plumbery, fittings=fakeFittings)
        self.facility.power_on()

    def tearDown(self):
        self.facility = None

    def test_get_location_id(self):
        self.assertEqual(self.facility.get_location_id(), 'NA9')

    def test_list_basement(self):
        self.assertEqual(self.facility.list_basement(), ['fake1'])

    def test_list_blueprints(self):
        self.assertEqual(self.facility.list_blueprints(), ['fake1', 'fake2'])

    def test_expand_blueprint(self):
        self.assertEqual(
            self.facility.expand_blueprint('fake1'), ['fake1'])
        self.assertEqual(
            self.facility.expand_blueprint('fake1 unknown fake'), ['fake1'])
        self.assertEqual(
            self.facility.expand_blueprint('macro'), ['fake2'])
        self.assertEqual(
            self.facility.expand_blueprint('basement'), ['fake1'])

    def test_get_blueprint(self):
        self.assertEqual(
            self.facility.get_blueprint('fake2')['target'], 'fake2')
        self.assertIsNone(self.facility.get_blueprint('macro'))
        self.assertIsNone(self.facility.get_blueprint('crazyAndunknown'))

    def test_list_domains(self):
        self.assertEqual(self.facility.list_domains(), ['VDC1'])

    def test_list_ethernets(self):
        self.assertEqual(self.facility.list_ethernets(), ['vlan1'])

    def test_list_nodes(self):
        self.assertEqual(self.facility.list_nodes(),
                         ['stackstorm1', 'stackstorm2'])

    def test_get_image(self):
        self.assertRegexpMatches(
            self.facility.get_image().name, "^RedHat ")
        self.assertIsNone(self.facility.get_image('perfectlyUnknown'))

    def test_focus(self):
        self.facility.focus()

    def test_process_all_blueprints(self):
        action = FakeAction({})
        self.facility.process_all_blueprints(action)
        self.assertEqual(action.count, 14)

    def test_process_blueprint(self):
        action = FakeAction({})
        self.facility.process_blueprint(action, names='fake')
        self.assertEqual(action.count, 4)

    def test_build_all_blueprints(self):
        self.facility.build_all_blueprints()

    def test_build_blueprint(self):
        self.facility.build_blueprint('fake')

    def test_start_all_blueprints(self):
        self.facility.start_all_blueprints()

    def test_start_blueprint(self):
        self.facility.start_blueprint('fake')

    def test_polish_all_blueprints(self):
        self.facility.polish_all_blueprints(polishers='ping')

    def test_polish_blueprint(self):
        self.facility.polish_blueprint(names='fake', polishers='ping')

    def test_stop_all_blueprints(self):
        self.facility.stop_all_blueprints()

    def test_stop_blueprint(self):
        self.facility.stop_blueprint('fake')

    def test_wipe_all_blueprints(self):
        self.facility.wipe_all_blueprints()

    def test_wipe_blueprint(self):
        self.facility.wipe_blueprint('fake')

    def test_destroy_all_blueprints(self):
        self.facility.destroy_all_blueprints()

    def test_destroy_blueprint(self):
        self.facility.destroy_blueprint('fake')

    def test_lookup(self):
        self.assertEqual(self.facility.lookup('location.country'), 'US')
        self.assertEqual(self.facility.lookup('location.id'), 'NA9')
        self.assertEqual(self.facility.lookup('*unknown*'), None)

    def test_settings(self):
        engine = PlumberyEngine(defaultsPlan)
        facility = engine.list_facility('EU6')[0]
        self.assertEqual(facility.get_setting('locationId'), 'EU6')
        self.assertEqual(facility.get_setting('regionId'), 'dd-eu')
        self.assertEqual(facility.get_setting('prepare'), None)
        self.assertEqual(facility.get_setting('basement'), 'myBlueprint')

    def test_settings_private(self):
        settings = {
            'apiHost': 'smee.com',
            'locationId': 'NA5',
            'safeMode': True
        }
        engine = PlumberyEngine(plan=settings)
        engine.set_user_name('smee')
        engine.set_user_password('smee')
        facility = engine.list_facility()[0]
#        facility.power_on()
        self.assertEqual(facility.get_setting('locationId'), 'NA5')
        self.assertIsNone(facility.get_setting('regionId'))

    def test_blueprints(self):
        engine = PlumberyEngine(defaultsPlan)
        facility = engine.list_facility('EU6')[0]
        blueprint = facility.get_blueprint('myBlueprint')
        self.assertEqual(isinstance(blueprint, dict), True)
        self.assertEqual(blueprint['domain']['name'], 'myDC')
        self.assertEqual(blueprint['domain']['ipv4'], 'auto')
        self.assertEqual(blueprint['ethernet']['name'], 'myVLAN')
        self.assertEqual(blueprint['ethernet']['subnet'], '10.1.10.0')
        self.assertEqual(len(blueprint['ethernet']['accept']), 1)
        label = list(blueprint['nodes'][0])[0]
        node = blueprint['nodes'][0][label]
        self.assertEqual('default' not in node, True)
        self.assertEqual(node['cpu'], 3)
        self.assertEqual(node['memory'], 5)
        self.assertEqual(len(node['information']), 2)
        config = node['cloud-config']
        self.assertEqual(isinstance(config, dict), True)
        self.assertEqual(config['disable_root'], True)
        self.assertEqual(config['ssh_pwauth'], True)
        self.assertEqual(len(config['packages']), 4)
        self.assertEqual(len(config['runcmd']), 4)

if __name__ == '__main__':
    import sys
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import time
from six import string_types
from plumbery.polishers.base import NodeConfiguration
from plumbery.exception import ConfigurationError
from plumbery.plogging import plogging


class BackupConfiguration(NodeConfiguration):
    __name__ = 'BackupConfiguration'
    _element_name_ = 'backup'
    _configuration_ = {
    }

    def __init__(self, engine, facility):
        self.engine = engine
        self.facility = facility

    def validate(self, settings):
        if self._element_name_ in settings:
            backup = settings[self._element_name_]
            if isinstance(backup, dict):
                plan = backup['plan'].lower().capitalize()
            else:
                plan = backup.lower().capitalize()
            if plan not in ['Essentials', 'Advanced', 'Enterprise']:
                raise ConfigurationError("Backup plan not valid")
            return True

    def configure(self, node, settings):
        if self._element_name_ in settings:
            self._configure_backup(node, settings[self._element_name_])
            return True
        return False

    def deconfigure(self, node, settings):
        return True

    def _configure_backup(self, node, backup):
        """
        Configure backup on a node

        :param node: the target node
        :type node: :class:`libcloud.compute.base.Node`

        :param backup: The backup settings
        :type  backup: ``dict`` or ``str``

        """
        default_email = self.facility.backup.connection.get_account_details().email
        if isinstance(backup, string_types):
            backup = {
                'plan': backup,
                'email': default_email,
                'clients': [{
                    'type': 'filesystem'
                }]
            }

        plan = backup['plan'].lower().capitalize()
        plogging.info("Starting {} backup of node '{}'".format(
            plan.lower(), node.name))

        backup_details = None
        try:
            self.facility.backup.create_target_from_node(
                node,
                extra={'servicePlan': plan})
        except Exception as feedback:
            if feedback.msg == 'Cloud backup for this server is already enabled or being enabled (state: NORMAL).':
                plogging.info("- already there")
                backup_details = self.facility.backup.ex_get_backup_details_for_target(node.id)
            else:
                plogging.info("- unable to start backup")
                plogging.error(str(feedback))
                return False

        while (backup_details is not None and
               backup_details.status is not 'NORMAL'):
            try:
                backup_details = self.facility.backup.ex_get_backup_details_for_target(node.id)
                plogging.info("- in progress, found asset %s", backup_details.asset_id)

            except Exception as feedback:
                if 'RESOURCE_BUSY' in str(feedback):
                    time.sleep(10)
                    continue

                elif 'RETRYABLE_SYSTEM_ERROR' in str(feedback):
                    time.sleep(10)
                    continue

                elif 'NO_CHANGE' in str(feedback):
                    plogging.info("- already there")

                elif 'RESOURCE_LOCKED' in str(feedback):
                    plogging.info("- unable to start backup "
                                 "- node has been locked")

                else:
                    plogging.info("- unable to start backup")
                    plogging.error(str(feedback))

            break
        target = self.facility.backup.ex_get_target_by_id(node.id)
        storage_policies = self.facility.backup.ex_list_available_storage_policies(
            target=target
        )
        schedule_policies = self.facility.backup.ex_list_available_schedule_policies(
            target=target
        )
        client_types = self.facility.backup.ex_list_available_client_types(
            target=target
        )
        clients = backup.get('clients', [{'type': 'filesystem'}])
        for client in clients:
            plogging.info("- adding backup client")

            client_type = client.get('type', 'filesystem').lower()
            storage_policy = client.get(
                'storagePolicy', '14 Day Storage Policy').lower()
            schedule_policy = client.get(
                'schedulePolicy', '12AM - 6AM').lower()
            trigger = client.get('trigger', 'ON_FAILURE')
            email = client.get('email', default_email)

            try:
                storage_policy = [x for x in storage_policies
                                  if x.name.lower() == storage_policy][0]
            except IndexError:
                raise ConfigurationError(
                    "Could not find matching storage policy '%s'" %
                    storage_policy)
            try:
                schedule_policy = [x for x in schedule_policies
                                   if x.name.lower() == schedule_policy][0]
            except IndexError:
                raise ConfigurationError(
                    "Could not find matching schedule policy '%s'" %
                    schedule_policy)

            if client_type in ['file', 'filesystem']:
                client = [x for x in client_types if x.is_file_system][0]
            else:
                client = [x for x in client_types
                          if x.description.startswith(client_type)][0]
            self.facility.backup.ex_add_client_to_target(
                target=target,
                client_type=client,
                storage_policy=storage_policy,
                schedule_policy=schedule_policy,
                trigger=trigger,
                email=email
            )
        return True


# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import copy
import socket
import os

from plumbery.action import PlumberyActionLoader
from plumbery.exception import PlumberyException
from plumbery.infrastructure import PlumberyInfrastructure
from plumbery.plogging import plogging
from plumbery.nodes import PlumberyNodes
from plumbery.polisher import PlumberyPolisher

__all__ = ['PlumberyFacility']


class PlumberyFacility(object):
    """
    Plumbing at one facility

    :param plumbery: the automate that is coordinating
            plumbing activities at multiple facilities
    :type plumbery: :class:`plumbery.PlumberyEngine`

    :param fittings:  the plan for the fittings
    :type fittings: ``dict``

    Example::

        from plumbery.facility import PlumberyFacility
        facility = PlumberyFacility(plumbery=plumbery, fittings=fittings)
        facility.build_all_blueprints()

    In this example a facility is ruled by the given plumber, and the plan
    of all the fittings is provided as well.

    Attributes:

        plumbery:
            global settings and functions

        fittings:
            the plan is available when needed

    """

    def __init__(self, plumbery=None, fittings={}):
        """Puts this facility in context"""

        self.plumbery = plumbery

        self.facts = {}

        self.settings = {}

        self.blueprints = []

        for key in fittings.keys():

            if key == 'blueprints':
                self.blueprints = fittings['blueprints']

            else:
                self.settings[key] = fittings[key]

        self.finalize_settings()
        self.finalize_blueprints()

        # first call to the API is done in self.power_on()
        self.region = None
        self.location = None
        self.backup = None

        self._cache_images = []
        self._cache_network_domains = []
        self._cache_vlans = []

    def __repr__(self):

        return "<PlumberyFacility settings: {}>".format(self.settings)

    def finalize_settings(self):
        """
        Sets values for settings

        This function gets missing settings from default values set at
        the engine level.
        """

        mandatory = []
        for label in mandatory:
            if label not in self.settings:
                value = self.plumbery.get_default(label)
                if value is None:
                    raise ValueError("No value has been set for '{}'"
                                     .format(label))
                self.settings[label] = value

        if 'regionId' not in self.settings and 'locationId' in self.settings:
            self.settings['regionId'] = self.get_region(
                self.settings['locationId'])

    def get_location_id(self):
        """
        Retrieves the id of the current location

        :return:  the id of the current location, e.g., 'EU7' or 'NA9'
        :rtype: ``str``

        """

        return self.get_setting('locationId')

    def get_region(self, locationId=None):
        """
        Retrieves the region for some location

        :param locationId: if not provided, use location of this facility
        :type locationId: ``str`` or None

        :return:  the region, e.g., 'dd-eu' or 'dd-ap', etc.
        :rtype: ``str`` or None

        This function helps to bind well-known data centres
        to their respective regions.

        For example::

            >>>facility.get_region(locationID='EU6')
            'dd-eu'

            >>>facility.get_region(locationID='XY7')
            None

        """

        regions = {
            'AF3': 'dd-af',
            'AP3': 'dd-ap',
            'AP4': 'dd-ap',
            'AP5': 'dd-ap',
            'AU9': 'dd-au',
            'AU10': 'dd-au',
            'AU11': 'dd-au',
            'CA2': 'dd-ca',
            'EU6': 'dd-eu',
            'EU7': 'dd-eu',
            'EU8': 'dd-eu',
            'EU9': 'dd-eu',
            'NA9': 'dd-na',
            'NA12': 'dd-na',
        }

        if locationId is None:
            locationId = self.get_setting('locationId')

        if locationId not in regions.keys():
            return None

        return regions[locationId]

    def get_setting(self, label, default=None):
        """
        Retrieves some setting

        :param label: the name of the setting to be retrieved
        :type label: ``str``

        :param default: the default value to return

        :return: the value set in fittings plan, or `None`
        :rtype: ``str`` most often

        """

        if label in self.settings:
            return self.settings[label]

        return self.plumbery.get_default(label, default)

    def list_basement(self):
        """
        Retrieves a list of blueprints that, together, constitute the basement
        of this facility

        :return: names of basement blueprints
        :rtype: ``list`` of ``str`` or ``[]``

        Basement is a list of blueprints defined in the fittings plan, as per
        following example::

            ---
            basement: admin control

            blueprints:

              - admin:
                  ...

        In that case you would get::

            >>facility.list_basement()
            ['admin', 'control']

        """

        basement = self.get_setting('basement')

        if basement is None:
            return []

        return self.expand_blueprint(basement)

    def finalize_blueprints(self):
        """
        Sets default values for blueprints

        This function expands blueprints defined for this facility with
        default settings and settings stored by the global engine.

        """

        for blueprintName in self.list_blueprints():

            blueprint = self.get_blueprint(blueprintName)

            shell = copy.deepcopy(self.plumbery.get_default('domain', {}))
            if 'domain' in blueprint:
                self.update_settings(shell, blueprint['domain'])
            blueprint['domain'] = shell

            shell = copy.deepcopy(self.plumbery.get_default('ethernet', {}))
            if 'ethernet' in blueprint:
                self.update_settings(shell, blueprint['ethernet'])
            blueprint['ethernet'] = shell

            if 'nodes' in blueprint:

                for index in range(0, len(blueprint['nodes'])):

                    config = copy.deepcopy(self.plumbery.get_default(
                        'cloud-config', {}))

                    if not isinstance(blueprint['nodes'][index], dict):
                        raise TypeError('{} should be a dictionary'
                                        .format(blueprint['nodes'][index]))

                    label = list(blueprint['nodes'][index])[0]
                    settings = blueprint['nodes'][index][label]
                    if settings is None:
                        settings = {}

                    shell = {}
                    if 'default' in settings:
                        shell = copy.deepcopy(self.plumbery.get_default(
                            settings['default'], {}))
                        settings.pop('default')
                        if 'cloud-config' in shell:
                            self.update_settings(config, shell['cloud-config'])
                            shell.pop('cloud-config')

                    if 'cloud-config' in settings:
                        self.update_settings(config, settings['cloud-config'])
                        settings.pop('cloud-config')

                    self.update_settings(shell, settings)
                    settings = shell

                    if len(list(config)) > 0:
                        settings['cloud-config'] = config

                    blueprint['nodes'][index][label] = settings

            for index in range(0, len(self.blueprints)):
                if list(self.blueprints[index])[0] == blueprintName:
                    self.blueprints[index][blueprintName] = blueprint

    def update_settings(self, settings, additions):
        """
        Updates a dictionary deeply

        :param settings: the dictionary that will be updated
        :type settings: ``dict``

        :param additions: dictionary with some updates
        :type additions: ``dict``

        This function appends to any list in settings the elements
        provided in the dictionary additions.

        """

        if not isinstance(additions, dict):
            return

        for key in additions.keys():
            if additions[key] is None:
                pass

            elif key not in settings:
                settings[key] = additions[key]

            elif isinstance(settings[key], list):
                settings[key] += additions[key]

            elif isinstance(settings[key], dict):
                self.update_settings(settings[key], additions[key])

            else:
                settings[key] = additions[key]

    def list_blueprints(self):
        """
        Retrieves a list of blueprints that have been defined

        :return: names of blueprints defined for this facility
        :rtype: ``list`` of ``str`` or ``[]``

        Blueprints are defined in the fittings plan, as per
        following example::

            ---
            blueprints:

              - admin:
                  ...

              - app: web data

              - web:
                  ...

              - data:
                  ...

        In that case you would get::

            >>facility.list_blueprints()
            ['admin', 'web', 'data']

        Note::
          As shown in the example above, only actionable blueprints are
          listed by this function.
        """

        names = []
        for blueprint in self.blueprints:
            name = list(blueprint)[0]
            if not isinstance(blueprint[name], dict):
                continue

            if name in names:
                raise ValueError("Duplicated blueprint name '{}'".format(name))

            names.append(name)

        return names

    def expand_blueprint(self, labels):
        """
        Designates multiple blueprints with a simple label

        :param labels: the label(s) to be expanded
        :type labels: ``str`` or ``list`` of ``str``

        :return: a list of names
        :rtype: ``list`` of ``str``

        Blueprints are defined in the fittings plan, as per
        following example::

            ---
            defaults:
              blueprints: mongo_config mongo_shard

            ---

            basement: control admin

            blueprints:

              - mongo: mongo_config mongo_mongos mongo_shard

              - mongo_config:
                  ...

              - mongo_mongos:
                  ...

              - mongo_shard:
                  ...

        In that case you would get::

            >>facility.expand_blueprint('mongo')
            ['mongo_config', 'mongo_mongos', 'mongo_shard']

            >>facility.expand_blueprint('mongo_config')
            ['mongo_config']

            >>facility.expand_blueprint('mongo_config mongo_mongos')
            ['mongo_config', 'mongo_mongos']

            >>facility.expand_blueprint(['mongo_config', 'alien_mongos'])
            ['mongo_config']

        If the special label `*` is provided, then plumbery will look
        for a default list of blueprints defined at the global level.
        Alternatively, it will provide the full list of blueprints.
        With the example settings above, you would get::

            >>facility.expand_blueprint('*')
            ['mongo_config', 'mongo_shard']

        If the special label `basement` is provided, then as you can expect
        it will be expanded as specified in the fittings plan. With the example
        settings above, you would get::

            >>facility.expand_blueprint('basement')
            ['control', 'admin']

        """

        names = []

        if isinstance(labels, str):

            if labels.lower() == '*':
                all = self.list_blueprints()
                labels = self.plumbery.get_default('blueprints', all)

            elif labels.lower() == 'basement':
                labels = self.get_setting('basement')

            if isinstance(labels, str):
                labels = labels.split(' ')

        for label in labels:

            for blueprint in self.blueprints:

                name = list(blueprint)[0]
                if name != label:
                    continue

                if isinstance(blueprint[name], dict):
                    if label not in names:
                        names.append(label)
                else:
                    for token in str(blueprint[name]).split(' '):
                        if token in names:
                            continue

                        for scanning in self.blueprints:
                            if token == list(scanning)[0]:
                                names.append(token)
                                break

        if names != labels:
            if len(names) > 1:
                plogging.info("- working on blueprints '{}'".format(
                    "', '".join(names)))
            elif len(names) > 0:
                plogging.info("- working on blueprint '{}'".format(
                    "', '".join(names)))
            else:
                plogging.info("- skipped - nothing to do here")

        return names

    def get_blueprint(self, name):
        """
        Retrieves a blueprint by name

        :param name: the name of the target blueprint
        :type name: ``str``

        :return: the blueprint with this name
        :rtype: ``dict`` or ``None``

        """

        for blueprint in self.blueprints:
            if name == list(blueprint)[0]:

                if not isinstance(blueprint[name], dict):
                    return None

                blueprint = blueprint[name]
                blueprint['target'] = name
                return blueprint

        return None

    def list_domains(self):
        """
        Retrieves the list of network domains that have been defined across
        blueprints for this facility

        :return: names of network domains defined for this facility
        :rtype: ``list`` of ``str`` or ``[]``

        Domains are defined in blueprints. Usually fittings plan are using at
        least one network domain, sometimes several.

        """

        labels = set()

        for blueprint in self.blueprints:
            name = list(blueprint)[0]
            if 'domain' in blueprint[name]:
                labels.add(blueprint[name]['domain']['name'])

        return list(labels)

    def list_ethernets(self):
        """
        Retrieves the list of Ethernet networks that have been defined across
        blueprints for this facility

        :return: names of Ethernet networks defined for this facility
        :rtype: ``list`` of ``str`` or ``[]``

        Ethernet networks are defined in blueprints. Usually fittings plan are
        using at least one Ethernet network, often several.

        """

        labels = set()

        for blueprint in self.blueprints:
            name = list(blueprint)[0]
            if 'ethernet' in blueprint[name]:
                labels.add(blueprint[name]['ethernet']['name'])

        return list(labels)

    def list_nodes(self):
        """
        Retrieves the list of nodes that have been defined across
        blueprints for this facility

        :return: names of nodes defined for this facility
        :rtype: ``list`` of ``str`` or ``[]``

        Nodes are defined in blueprints.

        """

        labels = []

        for blueprint in self.blueprints:
            name = list(blueprint)[0]
            if 'nodes' in blueprint[name]:
                for item in blueprint[name]['nodes']:
                    if type(item) is dict:
                        label = list(item)[0]

                    else:
                        label = item

                    for label in PlumberyNodes.expand_labels(label):
                        if label in labels:
                            plogging.warning("Duplicate node name '{}'"
                                             .format(label))
                        else:
                            labels.append(label)

        return labels

    def get_image(self, name=None):
        """
        Retrieves an acceptable image

        :param name: the name of the target image
        :type name: ``str``

        :return: a suitable image
        :rtype: :class:`Image` or ``None``

        This function looks at the library of available images and picks up
        the first image that has the name in it.

        Some examples::

            >>>facility.get_image('RedHat')
            ...
            >>>facility.get_image('RedHat 6 64-bit 4 CPU')
            ...

        """

        # cache images to limit API calls
        if len(self._cache_images) < 1:
            self.power_on()
            self._cache_images = \
                self.region.list_images(location=self.location)

            self._cache_images += \
                self.region.ex_list_customer_images(location=self.location)

        if (name is None and len(self._cache_images) > 0):
            return self._cache_images[0]

        for image in self._cache_images:
            if name in image.name:
                return image

        return None

    def focus(self):
        """
        Where are we plumbing?

        """

        self.power_on()
        plogging.info("Plumbing at '{}' {} ({})".format(
            self.location.id,
            self.location.name,
            self.location.country))

        blueprints = self.list_blueprints()
        if len(blueprints) < 1:
            plogging.warning("- no blueprint has been found")
        else:
            plogging.debug("- available blueprints: {}".format(
                "'"+"', '".join(blueprints)+"'"))

        basement = self.list_basement()
        if len(basement) > 0:
            plogging.debug("- basement: {}".format(
                "'"+"', '".join(basement)+"'"))

    def power_on(self):
        """
        Switches electricity on

        """

        regionId = self.get_setting('regionId')
        host = self.get_setting('apiHost')
        locationId = self.get_setting('locationId')

        try:

            if self.region is None:
                plogging.debug("Getting driver for '%s / %s'", regionId, host)
                self.region = self.plumbery.get_compute_driver(
                    region=regionId,
                    host=host)
                self.backup = self.plumbery.get_backup_driver(
                    region=regionId,
                    host=host)

                if os.getenv('LIBCLOUD_HTTP_PROXY') is not None:
                    plogging.debug('Setting proxy to %s' %
                                   (os.getenv('LIBCLOUD_HTTP_PROXY')))
                    self.region.connection.set_http_proxy(
                        proxy_url=os.getenv('LIBCLOUD_HTTP_PROXY'))
                    self.backup.connection.set_http_proxy(
                        proxy_url=os.getenv('LIBCLOUD_HTTP_PROXY'))
                    plogging.debug('Disabling SSL verification')
                    import libcloud.security
                    libcloud.security.VERIFY_SSL_CERT = False

            if self.location is None:
                plogging.debug("Getting location '{}'".format(locationId))
                locations = []
                for location in self.region.list_locations():
                    locations.append(location.id)
                    if location.id == locationId:
                        self.location = location

                if self.location is None:
                    plogging.info("Known locations: {}".format(locations))
                    raise PlumberyException("Unknown location '{}' in '{}'"
                                            .format(locationId, regionId))

        except ValueError:
            raise PlumberyException("Unknown region '{}'"
                                    .format(regionId))

        except socket.gaierror:
            raise PlumberyException("Cannot communicate with the API endpoint")

    def build_all_blueprints(self):
        """
        Builds all blueprints defined for this facility

        This function builds all network domains across all blueprints, then
        it builds all nodes across all blueprints.

        If the keyword ``basement`` mentions one or several blueprints,
        then these are built before the other blueprints.

        """

        self.power_on()
        infrastructure = PlumberyInfrastructure(self)
        nodes = PlumberyNodes(self)

        basement = self.list_basement()
        for name in basement:
            blueprint = self.get_blueprint(name)
            infrastructure.build(blueprint)

        blueprints = self.expand_blueprint('*')
        for name in blueprints:
            if name not in basement:
                blueprint = self.get_blueprint(name)
                infrastructure.build(blueprint)

        for name in basement:
            blueprint = self.get_blueprint(name)
            container = infrastructure.get_container(blueprint)
            nodes.build_blueprint(blueprint, container)

        for name in blueprints:
            if name not in basement:
                blueprint = self.get_blueprint(name)
                container = infrastructure.get_container(blueprint)
                nodes.build_blueprint(blueprint, container)

    def build_blueprint(self, names):
        """
        Builds a named blueprint for this facility

        :param names: the name(s) of the blueprint(s) to build
        :type names: ``str`` or ``list`` of ``str``

        This function builds the named blueprint in two steps: the
        infrastructure comes first, and then the nodes themselves.

            >>>facility.build_blueprint('sql')

        If the keyword ``basement`` mentions one or several blueprints,
        then network domains of these special blueprints are built before
        the actual target blueprint.

        Example ``fittings.yaml``::

            ---
            basement: admin

            blueprints:

              - admin:
                  ethernet: control

              - sql:
                  ethernet: data
                  nodes:
                    - server1:
                        glue: control

        In this example, the node ``server1``has two network interfaces. The
        main network interface is connected to the network ``data``, and the
        secondary network interface is connected to the network ``control``.

        """

        self.power_on()
        infrastructure = PlumberyInfrastructure(self)
        nodes = PlumberyNodes(self)

        basement = self.list_basement()
        for name in basement:
            blueprint = self.get_blueprint(name)
            infrastructure.build(blueprint)

        for name in self.expand_blueprint(names):

            blueprint = self.get_blueprint(name)

            if name not in basement:
                infrastructure.build(blueprint)

            nodes.build_blueprint(
                blueprint,
                infrastructure.get_container(blueprint))

    def process_all_blueprints(self, action):
        """
        Handles all blueprints at this facility

        :param action: the action to perform, e.g., 'start'
        :type action: ``str`` or :class:`plumbery.PlumberyAction`

        """

        if isinstance(action, str):
            action = PlumberyActionLoader.load(action)

        self.power_on()
        action.enter(self)

        basement = self.list_basement()
        for name in basement:
            blueprint = self.get_blueprint(name)
            action.process(blueprint)

        for name in self.expand_blueprint('*'):
            if name not in basement:
                blueprint = self.get_blueprint(name)
                action.process(blueprint)

        action.quit()

    def process_blueprint(self, action, names):
        """
        Handles one blueprint at this facility

        :param action: the action to perform, e.g., 'start'
        :type action: ``str`` or :class:`plumbery.PlumberyAction`

        :param names: the name(s) of the targeted blueprint(s)
        :type names: ``str`` or ``list`` of ``str``

        """

        if isinstance(action, str):
            action = PlumberyActionLoader.load(action)

        self.power_on()
        action.enter(self)

        for name in self.expand_blueprint(names):

            blueprint = self.get_blueprint(name)
            action.process(blueprint)

        action.quit()

    def start_all_blueprints(self):
        """
        Starts all nodes at this facility

        """

        basement = self.list_basement()
        for name in basement:
            self.start_blueprint(name)

        for name in self.expand_blueprint('*'):
            if name not in basement:
                self.start_blueprint(name)

    def start_blueprint(self, names):
        """
        Starts nodes from a given blueprint at this facility

        :param names: the name(s) of the target blueprint(s)
        :type names: ``str`` or ``list`` of ``str``

        """

        nodes = PlumberyNodes(self)

        for name in self.expand_blueprint(names):

            blueprint = self.get_blueprint(name)

            if 'nodes' not in blueprint:
                continue

            nodes.start_blueprint(blueprint)

    def polish_all_blueprints(self, polishers):
        """
        Walks all resources at this facility and polish them

        :param polishers: polishers to be applied
        :type polishers: list of :class:`plumbery.PlumberyPolisher`

        """

        basement = self.list_basement()
        for name in basement:
            plogging.debug("Processing blueprint '{}'".format(name))
            self.polish_blueprint(name, polishers)

        for name in self.expand_blueprint('*'):
            if name not in basement:
                plogging.debug("Processing blueprint '{}'".format(name))
                self.polish_blueprint(name, polishers)

    def polish_blueprint(self, names, polishers):
        """
        Walks a named blueprint for this facility and polish related resources

        :param names: the name(s) of the blueprint(s) to polish
        :type names: ``str`` or ``list`` of ``str``

        :param polishers: polishers to be applied
        :type polishers: list of :class:`plumbery.PlumberyPolisher`

        """

        if isinstance(polishers, str):
            polishers = PlumberyPolisher.filter(self.plumbery.polishers,
                                                polishers)

        self.power_on()
        infrastructure = PlumberyInfrastructure(self)
        nodes = PlumberyNodes(self)

        for polisher in polishers:
            polisher.move_to(self)

        for name in self.expand_blueprint(names):

            blueprint = self.get_blueprint(name)

            container = infrastructure.get_container(blueprint)

            for polisher in polishers:
                polisher.shine_container(container)

            nodes.polish_blueprint(blueprint, polishers, container)

    def stop_all_blueprints(self):
        """
        Stops all nodes at this facility

        """

        basement = self.list_basement()

        for name in self.expand_blueprint('*'):
            if name not in basement:
                self.stop_blueprint(name)

        for name in basement:
            self.stop_blueprint(name)

    def stop_blueprint(self, names):
        """
        Stops nodes of the given blueprint at this facility

        :param names: the name(s) of the target blueprint(s)
        :type names: ``str`` or ``list`` of ``str``

        You can use the following setting to prevent plumbery from stopping a
        node::

          - sql:
              domain: *vdc1
              ethernet: *data
              nodes:
                - slaveSQL:
                    running: always

        """

        nodes = PlumberyNodes(self)

        for name in self.expand_blueprint(names):

            blueprint = self.get_blueprint(name)

            if 'nodes' not in blueprint:
                continue

            nodes.stop_blueprint(blueprint)

    def wipe_all_blueprints(self):
        """
        Destroys all nodes at this facility

        """

        self.power_on()
        nodes = PlumberyNodes(self)

        basement = self.list_basement()

        for name in self.expand_blueprint('*'):
            if name in basement:
                continue
            blueprint = self.get_blueprint(name)
            plogging.debug("Wiping blueprint '{}'".format(name))
            nodes.destroy_blueprint(blueprint)

        for name in basement:
            blueprint = self.get_blueprint(name)
            plogging.debug("Wiping blueprint '{}'".format(name))
            nodes.destroy_blueprint(blueprint)

    def wipe_blueprint(self, names):
        """
        Destroys nodes of a given blueprint at this facility

        :param names: the names of the blueprint to destroy
        :type names: ``str`` or ``list`` of ``str``

        """

        self.power_on()
        nodes = PlumberyNodes(self)

        for name in self.expand_blueprint(names):

            blueprint = self.get_blueprint(name)
            nodes.destroy_blueprint(blueprint)

    def destroy_all_blueprints(self):
        """
        Destroys all blueprints at this facility

        """

        self.power_on()
        nodes = PlumberyNodes(self)
        infrastructure = PlumberyInfrastructure(self)

        basement = self.list_basement()

        for name in self.expand_blueprint('*'):
            if name in basement:
                continue
            blueprint = self.get_blueprint(name)
            plogging.debug("Destroying blueprint '{}'".format(name))
            nodes.destroy_blueprint(blueprint)
            infrastructure.destroy_blueprint(blueprint)

        for name in basement:
            blueprint = self.get_blueprint(name)
            plogging.debug("Destroying blueprint '{}'".format(name))
            nodes.destroy_blueprint(blueprint)
            infrastructure.destroy_blueprint(blueprint)

    def destroy_blueprint(self, names):
        """
        Destroys a given blueprint at this facility

        :param names: the name(s) of the blueprint(s) to destroy
        :type names: ``str`` or ``list`` of ``str``

        """

        self.power_on()
        nodes = PlumberyNodes(self)
        infrastructure = PlumberyInfrastructure(self)

        for name in self.expand_blueprint(names):

            blueprint = self.get_blueprint(name)
            nodes.destroy_blueprint(blueprint)
            infrastructure.destroy_blueprint(blueprint)

    def lookup(self, token):
        """
        Retrieves the value attached to a token

        :param token: the token, e.g., 'server1.ipv6'
        :type token: ``str``

        :return: the value attached to this token, or `None`

        """

        if token == 'location.country':
            return self.location.country

        if token == 'location.id':
            return self.get_location_id()

        if token in self.facts:
            return self.facts[token]

        return self.plumbery.lookup(token)

    def remember(self, token, value):
        """
        Remembers the value attached to a token

        :param token: the token, e.g., 'server1.private'
        :type token: ``str``

        :param value: the value attached to this token, e.g., '10.0.0.8'
        :type value: ``str``

        """

        self.facts[token] = value



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from plumbery.polishers.base import NodeConfiguration
from plumbery.exception import ConfigurationError

from plumbery.plogging import plogging


class MemoryConfiguration(NodeConfiguration):
    __name__ = 'MemoryConfiguration'
    _element_name_ = 'memory'
    _configuration_ = {
    }

    def validate(self, settings):
        if self._element_name_ in settings:
            memory = int(settings[self._element_name_])
            if memory < 1 or memory > 256:
                raise ConfigurationError("- memory should be between 1 and 256")

    def configure(self, node, settings):
        if self._element_name_ in settings:
            memory = int(settings[self._element_name_])
            plogging.debug("- setting {} GB of memory".format(
                    memory))
            return memory
        return False



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import hashlib
import os
import random
import requests
import string
import sys
import time
import uuid
import yaml
from six import string_types

from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization

from libcloud.compute.providers import get_driver as get_compute_driver
from libcloud.compute.types import Provider as ComputeProvider
from libcloud.loadbalancer.providers import get_driver as get_balancer_driver
from libcloud.loadbalancer.types import Provider as BalancerProvider
from libcloud.backup.providers import get_driver as get_backup_driver
from libcloud.backup.types import Provider as BackupProvider

from plumbery import __version__
from plumbery.action import PlumberyActionLoader
from plumbery.exception import PlumberyException
from plumbery.facility import PlumberyFacility
from plumbery.plogging import plogging
from plumbery.polisher import PlumberyPolisher
from plumbery.text import PlumberyText, PlumberyContext


__all__ = ['PlumberyEngine']


class PlumberyEngine(object):
    """
    Infrastructure as code at Dimension Data with Apache Libcloud

    :param plan: the fittings plan
    :type plan: ``str`` or ``file`` or ``dict`` or ``list`` of ``dict``

    :param parameters: the external parameters
    :type plan: ``str`` or ``file`` or ``dict``

    Plumbery is a convenient tool for infrastructure managers at cloud times.
    It allows for easy and repeatable deployments of various
    fittings, including compute nodes and related storage. It allows also for
    quick dismandling of the fittings when required.
    The idea here is not to compete with respected solutions such as chef or
    puppet. At its name implies, plumbery is targeting pipes and fittings, the
    very basic utility stuff that sophisticated tools can leverage.

    Example::

        from plumbery.engine import PlumberyEngine
        engine = PlumberyEngine('fittings.yaml')
        engine.build_all_blueprints()

    In this example the overall plan, in textual format, is given to the engine
    in preparation of subsequent processing. The software is not trying to
    guess a name by default, so if you do not provide a name, no configuration
    file is loaded. You can load the plan at any stage, or restart the engine
    with an updated plan, by invoking the member function ``parse_layout()``

    Note:
        While plumbery is not making assumptions for your configuration files,
        if your infrastructure is simple enough to fit in one single file then
        you are highly encouraged to name it ``fittings.yaml``

    Beyond the plan for your fittings, plumbery is also requiring some specific
    credentials to connect to cloud providers. To preserve the confidentiality
    of such information, it is read from the environment, and not from any
    configuration file. Therefore the need for local setup before running
    plumbery. This is part of the installation process.

    Last but not least, plumbery sets the root password of any new server that
    it creates. For obvious security reasons this is not taken from the
    fittings plan but from the environment, or it can be set in code.

    Under Linux, you may want to edit ``~/.bash_profile`` like this::

        # credentials to access cloud resources from Dimension Data
        export MCP_USERNAME='foo.bar'
        export MCP_PASSWORD='WhatsUpDoc'

        # password to access nodes remotely
        export SHARED_SECRET='*you really want to put a tricky password here*'

    These simple precautions are aiming to protect your infrastructure as much
    as possible. The general goal is to minimize risks induced by exposure to
    your fittings plan. You may lead transformation towards so-called
    infrastructure as code, and for this you will add version control to your
    fittings plan. As a result, plans will be stored in git or equivalent, and
    shared across some people.

    """

    def __init__(self, plan=None, parameters=None):
        """
        Ignites the plumbing engine

        :param plan: the fittings plan
        :type plan: ``str`` or ``file`` or ``dict`` or ``list`` of ``dict``

        :param parameters: the external parameters
        :type plan: ``str`` or ``file`` or ``dict``

        """
        self.c0 = time.clock()
        self.t0 = time.time()

        self.defaults = {}

        self.facilities = []

        self.information = []

        self.links = {}

        self.safeMode = False

        self.parameters = {}

        self.polishers = []

        self.buildPolisher = 'configure'

        self._sharedUser = None
        self._sharedSecret = None
        self._sharedKeyFiles = []

        self.secrets = {}
        self.secretsId = None

        self._userName = None

        self._userPassword = None

        # will be overridden to fittings path if provided
        self.working_directory = os.getcwd()
        self.set_parameters(parameters)
        self.set_fittings(plan)

    def set_parameters(self, parameters=None):
        """
        Changes the parameters of the running engine

        :param parameters: the new parameters
        :type parameters: ``str`` or ``file`` or ``dict``

        Parameters are a dictionary of key-value pairs.
        If a string is provided, it will contain a reference to some textual
        content (file name or URL).
        If a file handle is provided, content is read as YAML.

        """

        if not parameters:
            return

        if isinstance(parameters, string_types):

            if parameters.startswith(("https://", "http://")):
                response = requests.get(parameters)
                parameters = response.text
            else:
                parameters = open(parameters, 'r')
            parameters = yaml.load(parameters)

        if isinstance(parameters, list):
            pdict = {}
            for item in parameters:
                (key, value) = item.split('=')
                pdict[key] = value
            parameters = pdict

        plogging.debug("Parameters:")
        for key in parameters:
            if key not in self.parameters:
                self.parameters[key] = {}
            self.parameters[key]['value'] = parameters[key]
            plogging.debug("- {}: {}".format(
                key, self.parameters[key]['value']))

    def get_parameters(self):
        """
        Retrieves values of all parameters

        :return: a dictionary of all parameters
        :rtype: ``dict``

        """

        parameters = {}

        for key in self.parameters:
            parameters['parameter.'+key] = self.get_parameter(key)

        return parameters

    def get_parameter(self, label):
        """
        Retrieves value of some parameter

        :param label: the name of the parameter to be retrieved
        :type label: ``str``

        :return: the value set externallly, or default in fittings plan
        :rtype: ``str`` most often, or ``dict`` or something else

        """

        if label.startswith('parameter.'):
            label = label[len('parameter.'):]

        if label not in self.parameters:
            raise KeyError("Parameter '{}' is unknown".format(label))

        if 'value' in self.parameters[label]:
            return self.parameters[label]['value']

        if 'default' not in self.parameters[label]:
            raise ValueError("Parameter '{}' has no default value"
                             .format(label))

        return self.parameters[label]['default']

    def set_fittings(self, plan=None):
        """
        Loads the fittings plan for this engine instance

        :param plan: the file that contains fittings plan
        :type plan: ``str`` or ``file`` or ``dict`` or ``list`` of ``dict``

        The fittings plan is expected to follow YAML specifications, and it
        must have multiple documents in it. The first document provides
        general configuration parameters for the engine. Subsequent documents
        describe the various locations for the fittings.

        An example of a minimum fittings plan::

            # Frankfurt in Europe
            locationId: EU6

            blueprints:

              - myBluePrint:
                  domain:
                    name: myDC
                  ethernet:
                    name: myVLAN
                    subnet: 10.1.10.0
                  nodes:
                    - myServer

        In this example, the plan is to deploy a single node in the data centre
        at Frankfurt, in Europe. The node `myServer` will be placed in a
        network named `myVLAN`, and the network will be part of a network
        domain acting as a virtual data centre, `myDC`. The blueprint has a
        name, `myBluePrint`, so that it can be handled independently from
        other blueprints.

        """

        if plan is None:
            plan = os.getenv('PLUMBERY')

        if isinstance(plan, string_types):
            # hash reference to the fittings plan, not content of it
            self.secretsId = hashlib.md5(plan.encode('utf-8')).hexdigest()

            if plan.startswith(("https://", "http://")):
                response = requests.get(plan)
                plan = response.text

            elif plan == '-':
                plan = sys.stdin.read()

            elif '\n' not in plan:
                self.working_directory = os.path.dirname(plan)
                plan = open(plan, 'r').read()

            # load default values for parameters
            parameters = self.get_parameters()

            documents = plan.split('\n---')
            for document in documents:
                if '\n' in document:
                    settings = yaml.load(document)

                    if 'parameters' in settings:
                        for key in settings['parameters'].keys():
                            if 'parameter.'+key in parameters:
                                continue
                            if 'default' not in settings['parameters'][key]:
                                raise ValueError("Parameter '{}'"
                                                 " has no default value"
                                                 .format(key))
                            parameters['parameter.'+key] = settings['parameters'][key]['default']
                    break

            # expand parameters
            environment = PlumberyContext(dictionary=parameters)
            plan = PlumberyText.expand_parameters(plan, environment)

        if not plan:
            return

        if isinstance(plan, dict):
            documents = [plan]
            self.secretsId = hashlib.md5(str(plan).encode('utf-8')).hexdigest()

        elif isinstance(plan, list):
            documents = plan
            self.secretsId = hashlib.md5(str(plan).encode('utf-8')).hexdigest()

        else:
            documents = list(yaml.load_all(plan))

        if not isinstance(documents, list):
            raise TypeError('Fittings should be a list of dictionaries')

        # first document contains engine settings
        if len(documents) > 1:
            self.set_settings(documents.pop(0))

        # then one document per facility
        for document in documents:
            self.add_facility(document)

        self.load_secrets()

        if self.safeMode:
            plogging.info(
                "Running in safe mode"
                " - no actual change will be made to the fittings")

    def set_settings(self, settings):
        """
        Changes the settings of the engine

        :param settings: the new settings
        :type settings: ``dict``

        """

        if not isinstance(settings, dict):
            raise TypeError('settings should be a dictionary')

        if 'buildPolisher' in settings:
            if not isinstance(settings['buildPolisher'], str):
                raise TypeError('buildPolisher should be a string')

            self.buildPolisher = settings['buildPolisher']
            plogging.debug("Build polisher: {}".format(self.buildPolisher))

        if 'defaults' in settings:
            if not isinstance(settings['defaults'], dict):
                raise TypeError('defaults should be a dictionary')

            plogging.debug("Default values:")
            for key in settings['defaults'].keys():
                self.defaults[key] = settings['defaults'][key]
                plogging.debug("- {}: {}".format(key, self.defaults[key]))

        if 'information' in settings:
            if isinstance(settings['information'], str):
                settings['information'] = \
                    settings['information'].strip('\n').split('\n')
            if not isinstance(settings['information'], list):
                raise TypeError('information should be a list')

            self.information = settings['information']

        if 'keys' in settings:
            if not isinstance(settings['keys'], list):
                raise TypeError('keys should be a list')

            self.set_shared_key_files(settings['keys'])

        else:
            self.set_shared_key_files()

        if 'links' in settings:
            if not isinstance(settings['links'], dict):
                raise TypeError('links should be a dictionary')

            self.links = settings['links']

        if 'parameters' in settings:
            if not isinstance(settings['parameters'], dict):
                raise TypeError('parameters should be a dictionary')

            plogging.debug("Parameters:")
            for key in settings['parameters']:

                if key not in self.parameters:
                    self.parameters[key] = {}

                if 'information' not in settings['parameters'][key]:
                    raise ValueError("Parameter '{}' has no information"
                                     .format(key))
                self.parameters[key]['information'] = \
                    settings['parameters'][key]['information']

                if 'type' not in settings['parameters'][key]:
                    raise ValueError("Parameter '{}' has no type"
                                     .format(key))
                self.parameters[key]['type'] = \
                    settings['parameters'][key]['type']

                if 'default' not in settings['parameters'][key]:
                    plogging.warning("Parameter '{}' has no default value"
                                     .format(key))
                else:
                    self.parameters[key]['default'] = \
                        settings['parameters'][key]['default']
                    plogging.debug("- {}: {}".format(
                        key,
                        self.parameters[key]['default']))

        if 'actions' in settings:
            if not isinstance(settings['actions'], list):
                raise TypeError('actions should be a list')

            plogging.debug("Actions:")
            for item in settings['actions']:
                key = list(item)[0]
                value = item[key]
                self.polishers.append(
                    PlumberyPolisher.from_shelf(key, value))
                plogging.debug("- {}".format(key))

        if 'safeMode' in settings:
            if settings['safeMode'] not in [True, False]:
                raise ValueError('safeMode should be either True or False')
            self.safeMode = settings['safeMode']

    def get_default(self, label, default=None):
        """
        Retrieves default settings

        :param label: the name of the settings to be retrieved
        :type label: ``str``

        :param default: the default value to return

        :return: the value set in fittings plan, or `None`
        :rtype: ``dict`` most often, or ``str`` or something else

        """

        if label in self.defaults:
            return self.defaults[label]

        return default

    def set_shared_user(self, user):
        """
        Changes the name used to access nodes remotely

        :param user: the user name to be used with ssh
        :type user: ``str``

        """

        self._sharedUser = user

    def get_shared_user(self):
        """
        Retrieves the name used to access nodes remotely

        :return: the user name to be used with ssh
        :rtype: ``str``

        This functions returns ``root`` until the member function
        ``set_shared_used()`` has been called.

        """

        if self._sharedUser is None:
            return 'root'

        return self._sharedUser

    def set_shared_secret(self, secret):
        """
        Changes the shared secret to be used with new nodes

        :param secret: the shared secret to be used with the API and with ssh
        :type secret: ``str``

        This function can be used to supplement the normal provision of
        the shared secret via the environment variable ``SHARED_SECRET``.

        """

        self._sharedSecret = secret

    def get_shared_secret(self):
        """
        Retrieves the secret that is communicated to new nodes during setup

        :return: the shared secret to be used with the API and with ssh
        :rtype: ``str``

        :raises: :class:`plumbery.PlumberyException`
            - if no shared secret can be found

        The shared secret is not put in the fittings plan, but is normally
        taken from the environment variable ``SHARED_SECRET``.

        Under Linux, you may want to edit ``~/.bash_profile`` like this::

            # password to access nodes remotely
            export SHARED_SECRET='*you really want to use a tricky password*'

        Alternatively, you can use the member function ``set_shared_secret()``
        to set this important attribute via code.

        """

        if self._sharedSecret is None:
            self._sharedSecret = os.getenv('SHARED_SECRET')
            if self._sharedSecret is None or len(self._sharedSecret) < 3:
                raise PlumberyException(
                    "Error: missing node password "
                    "in environment SHARED_SECRET")

        return self._sharedSecret

    def set_shared_key_files(self, keys=None):
        """
        Sets locations of ssh keys

        :param keys: additional path name that contains a ssh key
        :type keys: ``str`` or ``list`` of ``str`` or `None`

        This function can be used to complement the normal provision of
        ssh keys, or to check that some key is available.

        The functions looks for keys in the
        user context, and from the environment.

        A ValueError exception is raised if no key can be found, or if the
        argument provided is not an existing file.

        """

        self._sharedKeyFiles = self.get_shared_key_files()

        if keys is not None:
            if isinstance(keys, str):
                keys = [keys]
            for key in keys:
                file = os.path.expanduser(key)
                if not os.path.isfile(file):
                    raise ValueError("Error: missing file {}".format(key))
                if key not in self._sharedKeyFiles:
                    plogging.debug("- using shared key {}".format(key))
                    self._sharedKeyFiles.insert(0, key)

        if len(self._sharedKeyFiles) < 1:
            plogging.error(
                "Error: no SSH key could be found, please run "
                "ssh-keygen -t rsa -b 4096 -C <your email@c.com>")

    def get_shared_key_files(self):
        """
        Retrieves locations of ssh keys

        :return: path names for files that contain ssh keys
        :rtype: ``list`` of ``str`` or `None`

        SSH keys are not put in the fittings plan, but normally
        taken from user environment, and from the variable ``SHARED_KEY``.

        You can use the member function ``set_shared_key_files()`` to add
        a key to the default set of keys.

        """

        if len(self._sharedKeyFiles) > 0:
            return self._sharedKeyFiles

        key = os.getenv('SHARED_KEY')
        if key is not None:
            file = os.path.expanduser(key)
            if not os.path.isfile(file):
                raise ValueError(
                    "Error: non-existent file "
                    "SHARED_KEY={}".format(key))
            plogging.debug("- using shared key {}".format(key))
            self._sharedKeyFiles = [key]

        # from http://www.programcreek.com/python/example/5607/paramiko.RSAKey
        #
        for key in ('~/.ssh/id_rsa.pub', # Unix
                    '~/.ssh/id_dsa.pub', # Unix
                    '~/ssh/id_rsa.pub',  # Windows
                    '~/ssh/id_dsa.pub'): # Windows
            file = os.path.expanduser(key)
            if os.path.isfile(file):
                plogging.debug("- using shared key {}".format(key))
                self._sharedKeyFiles.append(key)

        return self._sharedKeyFiles

    def get_rsa_secret(self, id='rsa_private.pair'):
        """
        Returns a part of a RSA pair of keys

        :param id: name of the key
        :type id: ``str``

        """

        type = '.'.join([x for x in id.split('.')
                        if x in ('rsa_private', 'rsa_public')])

        name = '.'.join([x for x in id.split('.')
                        if x not in ('rsa_private', 'rsa_public')])

        id = type+'.'+name

        if id in self.secrets:
            return self.secrets[id]

        if id == 'rsa_private.local':
            raise LookupError("It is forbidden to use 'rsa_private.local'")

        if id == 'rsa_public.local':
            paths = self.get_shared_key_files()
            if len(paths) < 1:
                return ''
            path = paths.pop(0)

            try:
                with open(os.path.expanduser(path)) as stream:
                    plogging.debug("- loading {} from {}".format(id, path))
                    text = stream.read().strip()
                    stream.close()
                    return text

            except IOError:
                plogging.error("- cannot load {} from {}".format(id, path))
                return ''

        key = rsa.generate_private_key(
            public_exponent=65537,
            key_size=2048,
            backend=default_backend())
        self.secrets['rsa_private.'+name] = key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption())
        plogging.debug("- generating {}".format('rsa_private.'+name))

        pubkey = key.public_key()
        self.secrets['rsa_public.'+name] = pubkey.public_bytes(
            encoding=serialization.Encoding.OpenSSH,
            format=serialization.PublicFormat.OpenSSH)
        plogging.debug("- generating {}".format('rsa_public.'+name))

        self.save_secrets()
        return self.secrets[id]

    def get_secret(self, id='secret.random'):
        """
        Returns a secret

        :param id: name of this secret
        :type id: ``str``

        :return: a transient random secret
        :rtype: ``str``

        Random secrets can be used in scripts and in configuration files
        sent to nodes, for example to configure a database server.

        For this you would put ``{{ secret.random }}`` in your files and let
        plumbery provide a value for you.

        The `id` parameter designates one secret among several.

        By default this function builds a random string out of ASCII letters,
        digits, and a couple of punctuation letters, with 9 characters.

        If you put `.uuid` in the id, than the function ``uuid.uuid4()`` is
        called to generate a unique identifier of 36 letters.

        The `format` parameter specifies the kind of string that is expected:
        * 'sha1' - rather long string
        * 'md5' - rather long string too
        * 'sha256' - longest and most secure

        Some examples:

            {{ uuid.server1 }}
            {{ uuid.server2 }}
            {{ secret.sql_root }}
            {{ secret.redis.master.md5 }}
            {{ secret.redis.slave }}
            {{ secret.database357.sha1 }}

        In a nutshell, this function is giving you a lot of flexibility
        in the generation of secrets.

        """

        if id in self.secrets:
            return self.secrets[id]

        if id.startswith('uuid.') or id.endswith('.uuid'):
            secret = str(uuid.uuid4())

        elif id.startswith('http://') or id.startswith('https://'):
            plogging.debug('- fetching {}'.format(id))
            secret = requests.get(id).text

        else:
            secret = ''.join(random.choice(
                string.ascii_letters+string.digits+'-_!=')
                for i in range(9))

        if id.endswith('.sha256'):
            secret = hashlib.sha256(secret.encode('utf-8')).hexdigest()

        elif id.endswith('.md5'):
            secret = hashlib.md5(secret.encode('utf-8')).hexdigest()

        elif id.endswith('.sha1'):
            secret = hashlib.sha1(secret.encode('utf-8')).hexdigest()

        plogging.debug("- generating {}".format(id))
        self.secrets[id] = secret
        self.save_secrets()

        return secret

    def display_secrets(self):
        """
        Displays secrets attached to this fittings plan
        """

        plogging.info("Showing secrets")

        if len(self.secrets.keys()) < 1:
            plogging.info("- no secret found")

        for key in sorted(self.secrets):
            plogging.info("- {}: {}".format(key, self.secrets[key]))

    def save_secrets(self, plan=None):
        """
        Saves secrets attached to this fittings plan

        :param plan: the file that contains fittings plan
        :type plan: ``str`` or ``file`` or ``dict`` or ``list`` of ``dict``

        """

        if plan:
            secretsId = hashlib.md5(plan.encode('utf-8')).hexdigest()

        elif self.secretsId:
            secretsId = self.secretsId

        else:
            return

        secretsFile = secretsId+'.secrets'

        try:
            handle = open(secretsFile, 'w')
            for id in self.secrets:
                handle.write("{}: '{}'\n".format(
                    id, self.secrets[id].replace('\n', '\\n')))
            handle.close()

        except IOError:
            plogging.warning("Unable to save secrets")
            plogging.debug("- cannot write to file '{}'".format(
                secretsFile))

    def load_secrets(self, plan=None):
        """
        Loads secrets attached to this fittings plan

        :param plan: the file that contains fittings plan
        :type plan: ``str`` or ``file`` or ``dict`` or ``list`` of ``dict``

        """

        if plan:
            secretsId = hashlib.md5(plan.encode('utf-8')).hexdigest()

        elif self.secretsId:
            secretsId = self.secretsId

        else:
            return

        secretsFile = secretsId+'.secrets'
        plogging.debug("Loading secrets from '{}'".format(secretsFile))

        if os.path.isfile(secretsFile):
            try:
                handle = open(secretsFile, 'r')
                self.secrets = yaml.load(handle)
                handle.close()

                plogging.debug("- found {} secrets".format(
                    len(self.secrets)))

            except IOError:
                plogging.debug("- unable to load secrets")

    def forget_secrets(self, plan=None):
        """
        Destroys secrets attached to this fittings plan
        """

        if plan:
            secretsId = hashlib.md5(plan.encode('utf-8')).hexdigest()

        elif self.secretsId:
            secretsId = self.secretsId

        else:
            return

        secretsFile = secretsId+'.secrets'

        if self.safeMode:
            plogging.info("Secrets cannot be forgotten in safe mode")

        self.secrets = {}

        if os.path.isfile(secretsFile):
            try:
                os.remove(secretsFile)

            except IOError:
                plogging.warning("Unable to forget secrets")
                plogging.debug("- cannot delete file '{}'".format(
                    secretsFile))

    def set_user_name(self, name):
        """
        Changes the name used to authenticate to the API

        :param name: the user name to be used with the driver
        :type name: ``str``

        This function can be used to supplement the normal provision of
        a user name via the environment variable ``MCP_USERNAME``.

        """

        self._userName = name

    def get_user_name(self):
        """
        Retrieves user name to authenticate to the API

        :return: the user name to be used with the driver
        :rtype: ``str``

        :raises: :class:`plumbery.PlumberyException`
            - if no user name can be found

        The user name is not put in the fittings plan, but is normally taken
        from the environment variable ``MCP_USERNAME``.

        Under Linux, you may want to edit ``~/.bash_profile`` like this::

            # credentials to access cloud resources from Dimension Data
            export MCP_USERNAME='foo.bar'
            export MCP_PASSWORD='WhatsUpDoc'

        """

        if self._userName is None:
            self._userName = os.getenv('MCP_USERNAME')
            if self._userName is None or len(self._userName) < 3:
                raise PlumberyException(
                    "Error: missing credentials in environment MCP_USERNAME")

        return self._userName

    def set_user_password(self, password):
        """
        Changes the password used to authenticate to the API

        :param password: the user password to be used with the driver
        :type password: ``str``

        This function can be used to supplement the normal provision of
        a user password via the environment variable ``MCP_PASSWORD``.

        """

        self._userPassword = password

    def get_user_password(self):
        """
        Retrieves user password to authenticate to the API

        :return: the user password to be used with the driver
        :rtype: ``str``

        :raises: :class:`plumbery.PlumberyException`
            - if no user password can be found

        The user password is not put in the fittings plan, but is normally
        taken from the environment variable ``MCP_PASSWORD``.

        Under Linux, you may want to edit ``~/.bash_profile`` like this::

            # credentials to access cloud resources from Dimension Data
            export MCP_USERNAME='foo.bar'
            export MCP_PASSWORD='WhatsUpDoc'

        """

        if self._userPassword is None:
            self._userPassword = os.getenv('MCP_PASSWORD')
            if self._userPassword is None or len(self._userPassword) < 3:
                raise PlumberyException(
                    "Error: missing credentials in environment MCP_PASSWORD")

        return self._userPassword

    def add_facility(self, facility):
        """
        Extends the scope of this plumbing engine

        :param facility: description of an additional facility
        :type facility: ``dict`` or :class:`plumbery.PlumberyFacility`

        """

        if isinstance(facility, dict):
            facility = PlumberyFacility(self, facility)

        plogging.debug("Adding facility %s" % facility)
        for key in facility.settings:
            plogging.debug("- {}: {}".format(key, facility.settings[key]))

        self.facilities.append(facility)

    def list_facility(self, location=None):
        """
        Retrieves facilities by their location

        :param location: the target location, e.g., 'EU6'
        :type location: ``str`` or ``list`` of ``str``

        :return: the list of matching facilities
        :rtype: ``list`` of :class:`plumbery.PlumberyFacility` or ``[]``

        Examples::

            >>>engine.list_facility(location='EU6')
            ...

            >>>engine.list_facility(location='EU6 NA9')
            ...

            >>>engine.list_facility(location=['EU6', 'NA9'])
            ...

        """
        if location is None:
            return self.facilities

        if isinstance(location, str):
            location = location.split(' ')

        matches = []

        for item in location:
            if isinstance(item, PlumberyFacility):
                matches.append(item)

        for facility in self.facilities:
            if (facility.get_setting('locationId') in location or
                    facility.get_setting('apiHost') in location):
                matches.append(facility)

        return matches

    def document_elapsed(self, elapsed=None, www=None):
        """
        Reflects the amount of work done by plumbery

        :param elapsed: number of second spent in plumbing
        :type elapsed: ``int``

        :param www: number of second spent in the cloud
        :type www: ``int``

        """

        if elapsed is None:
            elapsed = int(time.clock() - self.c0 + 1)

        if www is None:
            www = int(time.time() - self.t0 + 1)

        www -= elapsed

        if elapsed < 2:
            elapsed = "in a blink"

        elif elapsed < 120:
            elapsed = "for {} seconds".format(elapsed)

        elif elapsed < 7200:
            elapsed = int(elapsed/60)
            elapsed = "for {} minutes".format(elapsed)

        elif elapsed < 86400:
            elapsed = int(elapsed/60)
            minutes = elapsed % 60
            hours = int(elapsed / 60)
            elapsed = "for {} hours and {} minutes".format(hours, minutes)

        if www < 120:
            www = "for {} seconds".format(www)

        elif www < 7200:
            www = int(www/60)
            www = "for {} minutes".format(www)

        elif www < 86400:
            www = int(www/60)
            minutes = www % 60
            hours = int(www/60)
            www = "for {} hours and {} minutes".format(hours, minutes)

        return "Worked for you {} locally, and {} in the cloud"     \
               .format(elapsed, www)

    def do(self, action, blueprints=None, facilities=None):
        """
        Applies an action to multiple blueprints at multiple locations

        :param action: 'deploy', 'dispose', etc., or a polisher name
        :type action: ``str``

        :param blueprints: a list of blueprints
        :type blueprints: ``list`` of ``str``, or None

        :param facilities: a list of facilities
        :type factilities: ``list`` of ``str``, or None

        """

        if action == 'build':
            if blueprints is None:
                self.build_all_blueprints(facilities)
            else:
                self.build_blueprint(blueprints, facilities)

        elif action == 'deploy':
            if blueprints is None:
                self.build_all_blueprints(facilities)
                self.start_all_blueprints(facilities)
                self.polish_all_blueprints(filter='prepare',
                                           facilities=facilities)
                self.polish_all_blueprints(filter='information',
                                           facilities=facilities)
            else:
                self.build_blueprint(blueprints, facilities)
                self.start_blueprint(blueprints, facilities)
                self.polish_blueprint(blueprints,
                                      filter='prepare',
                                      facilities=facilities)
                self.polish_blueprint(blueprints,
                                      filter='information',
                                      facilities=facilities)

        elif action == 'destroy':
            if blueprints is None:
                self.destroy_all_blueprints(facilities)
            else:
                self.destroy_blueprint(blueprints, facilities)

        elif action == 'dispose':
            if blueprints is None:
                self.stop_all_blueprints(facilities)
                self.destroy_all_blueprints(facilities)
            else:
                self.stop_blueprint(blueprints, facilities)
                self.destroy_blueprint(blueprints, facilities)

        elif action in ('polish', 'finalize', 'finalise'):
            if blueprints is None:
                self.polish_all_blueprints(filter=None,
                                           facilities=facilities)
            else:
                self.polish_blueprint(blueprints,
                                      filter=None,
                                      facilities=facilities)

        elif action == 'refresh':
            if blueprints is None:
                self.stop_all_blueprints(facilities)
                self.wipe_all_blueprints(facilities)
                time.sleep(30)
                self.build_all_blueprints(facilities)
                self.start_all_blueprints(facilities)
                self.polish_all_blueprints(filter='prepare',
                                           facilities=facilities)
                self.polish_all_blueprints(filter='information',
                                           facilities=facilities)
            else:
                self.stop_blueprint(blueprints, facilities)
                self.wipe_blueprint(blueprints, facilities)
                time.sleep(30)
                self.build_blueprint(blueprints, facilities)
                self.start_blueprint(blueprints, facilities)
                self.polish_blueprint(blueprints,
                                      filter='prepare',
                                      facilities=facilities)
                self.polish_blueprint(blueprints,
                                      filter='information',
                                      facilities=facilities)

        elif action == 'secrets':
            self.display_secrets()

        elif action == 'start':
            if blueprints is None:
                self.start_all_blueprints(facilities)
            else:
                self.start_blueprint(blueprints, facilities)

        elif action == 'stop':
            if blueprints is None:
                self.stop_all_blueprints(facilities)
            else:
                self.stop_blueprint(blueprints, facilities)

        elif action == 'wipe':
            if blueprints is None:
                self.wipe_all_blueprints(facilities)
            else:
                self.wipe_blueprint(blueprints, facilities)

        elif action == 'graph':
            raise NotImplementedError("Wait for 1.1.0")

        else:
            if blueprints is None:
                self.polish_all_blueprints(filter=action,
                                           facilities=facilities)
            else:
                self.polish_blueprint(blueprints,
                                      filter=action,
                                      facilities=facilities)

    def process_all_blueprints(self, action, facilities=None):
        """
        Handles elements described in the fittings plan

        :param action: the action to perform, e.g., 'start'
        :type action: ``str`` or :class:`plumbery.PlumberyAction`

        :param facilities: explicit list of target facilities
        :type facilities: ``str`` or ``list`` of ``str``

        This function triggers the provided action across facilities.
        The default behaviour is to consider all facilities mentioned in the
        fittings plan. If a list of facilities is provided, than the action is
        limited to this subset only.

        """

        if isinstance(action, str):
            action = PlumberyActionLoader.load(action)

        action.begin(self)

        if facilities is not None:
            facilities = self.list_facility(facilities)
        else:
            facilities = self.facilities

        for facility in facilities:
            facility.focus()
            facility.process_all_blueprints(action)

        action.end()

    def process_blueprint(self, action, names, facilities=None):
        """
        Handles elements for one blueprint of the fittings plan

        :param action: the action to perform, e.g., 'start'
        :type action: ``str`` or :class:`plumbery.PlumberyAction`

        :param names: the name(s) of the targeted blueprint(s)
        :type names: ``str`` or ``list`` of ``str``

        :param facilities: explicit list of target facilities
        :type facilities: ``str`` or ``list`` of ``str``

        This function triggers the provided action across facilities.
        The default behaviour is to consider all facilities mentioned in the
        fittings plan. If a list of facilities is provided, than the action is
        limited to this subset only.

        """

        if isinstance(action, str):
            action = PlumberyActionLoader.load(action)

        action.begin(self)

        if isinstance(names, list):
            names = ' '.join(names)

        if facilities is not None:
            facilities = self.list_facility(facilities)
        else:
            facilities = self.facilities

        for facility in facilities:
            facility.focus()
            facility.process_blueprint(action, names)

        action.end()

    def build_all_blueprints(self, facilities=None):
        """
        Builds all blueprints described in fittings plan

        :param facilities: explicit list of target facilities
        :type facilities: ``str`` or ``list`` of ``str``

        This function checks facilities to build all blueprints there.
        The default behaviour is to consider all facilities mentioned in the
        fittings plan. If a list of facilities is provided, than the action is
        limited to this subset only.

        Examples::

            from plumbery.engine import PlumberyEngine
            PlumberyEngine('fittings.yaml').build_all_blueprints()

            from plumbery.engine import PlumberyEngine
            PlumberyEngine('fittings.yaml').build_all_blueprints('EU6 NA9')

        """

        all = self.get_default('blueprints', None)
        if all is None:
            plogging.info("Building all blueprints")
        else:
            plogging.info("Building '{}'".format("', '".join(all.split(' '))))

        if facilities is not None:
            facilities = self.list_facility(facilities)
        else:
            facilities = self.facilities

        for facility in facilities:
            facility.focus()
            facility.build_all_blueprints()

        self.polish_all_blueprints(filter=self.buildPolisher,
                                   facilities=facilities)

    def build_blueprint(self, names, facilities=None):
        """
        Builds named blueprint from fittings plan

        :param names: the name(s) of the blueprint(s) to deploy
        :type names: ``str`` or ``list`` of ``str``

        :param facilities: explicit list of target facilities
        :type facilities: ``str`` or ``list`` of ``str``

        This function checks facilities to build one single blueprint there.
        The default behaviour is to consider all facilities mentioned in the
        fittings plan. If a list of facilities is provided, than the action is
        limited to this subset only.

        Example::

            from plumbery.engine import PlumberyEngine
            PlumberyEngine('fittings.yaml').build_blueprints('sql')

        """

        if isinstance(names, list):
            label = ' '.join(names)
        else:
            label = names

        plogging.info("Building blueprint '{}'".format(label))

        if facilities is not None:
            facilities = self.list_facility(facilities)
        else:
            facilities = self.facilities

        for facility in facilities:
            facility.focus()
            facility.build_blueprint(names)

        self.polish_blueprint(names=names,
                              filter=self.buildPolisher,
                              facilities=facilities)

    def start_all_blueprints(self, facilities=None):
        """
        Starts all nodes described in the fittings plan

        :param facilities: explicit list of target facilities
        :type facilities: ``str`` or ``list`` of ``str``

        This function checks facilities to start all nodes there.
        The default behaviour is to consider all facilities mentioned in the
        fittings plan. If a list of facilities is provided, than the action is
        limited to this subset only.

        This function has no effect on nodes that are already up and running.

        """

        plogging.info("Starting nodes from all blueprints")

        if facilities is not None:
            facilities = self.list_facility(facilities)
        else:
            facilities = self.facilities

        for facility in facilities:
            facility.focus()
            facility.start_all_blueprints()

    def start_blueprint(self, names, facilities=None):
        """
        Starts nodes of one blueprint of the fittings plan

        :param names: the name(s) of the blueprint(s) to start
        :type names: ``str`` or ``list`` of ``str``

        :param facilities: explicit list of target facilities
        :type facilities: ``str`` or ``list`` of ``str``

        This function checks facilities to start nodes from some blueprint.
        The default behaviour is to consider all facilities mentioned in the
        fittings plan. If a list of facilities is provided, than the action is
        limited to this subset only.

        This function has no effect on nodes that are already up and running.

        """

        if isinstance(names, list):
            label = ' '.join(names)
        else:
            label = names

        plogging.info("Starting nodes from blueprint '{}'".format(label))

        if facilities is not None:
            facilities = self.list_facility(facilities)
        else:
            facilities = self.facilities

        for facility in facilities:
            facility.focus()
            facility.start_blueprint(names)

    def polish_all_blueprints(self, filter=None, facilities=None):
        """
        Walks all resources and polishes them

        :param filter: the name of a single polisher to apply. If this
            parameter is missing, all polishers declared in the fittings plan
            will be applied
        :type filter: ``str``

        :param facilities: explicit list of target facilities
        :type facilities: ``str`` or ``list`` of ``str``

        This function checks facilities to apply polishers there.
        The default behaviour is to consider all facilities mentioned in the
        fittings plan. If a list of facilities is provided, than the action is
        limited to this subset only.

        Example::

            from plumbery.engine import PlumberyEngine
            PlumberyEngine('fittings.yaml').polish_all_blueprints()

        """

        polishers = PlumberyPolisher.filter(self.polishers, filter)

        if len(polishers) < 1:
            return False

        plogging.info("Processing all blueprints")

        for polisher in polishers:
            polisher.go(self)

        if facilities is not None:
            facilities = self.list_facility(facilities)
        else:
            facilities = self.facilities

        for facility in facilities:
            facility.focus()
            facility.polish_all_blueprints(polishers)

        for polisher in polishers:
            polisher.reap()

    def polish_blueprint(self, names, filter=None, facilities=None):
        """
        Walks resources from the target blueprint and polishes them

        :param names: the name(s) of the blueprint(s) to polish
        :type names: ``str`` or ``list`` of ``str``

        :param filter: the name of a single polisher to apply. If this
            parameter is missing, all polishers declared in the fittings plan
            will be applied
        :type filter: ``str``

        :param facilities: explicit list of target facilities
        :type facilities: ``str`` or ``list`` of ``str``

        This function checks facilities to apply one polisher to one blueprint.
        The default behaviour is to consider all facilities mentioned in the
        fittings plan. If a list of facilities is provided, than the action is
        limited to this subset only.

        Example::

            from plumbery.engine import PlumberyEngine
            PlumberyEngine('fittings.yaml').polish_blueprint('sql')

        """

        polishers = PlumberyPolisher.filter(self.polishers, filter)

        if len(polishers) < 1:
            plogging.debug('No polisher has been found')
            return

        if isinstance(names, list):
            label = ' '.join(names)
        else:
            label = names

        plogging.info("Processing blueprint '{}'".format(label))

        for polisher in polishers:
            polisher.go(self)

        if facilities is not None:
            facilities = self.list_facility(facilities)
        else:
            facilities = self.facilities

        for facility in facilities:
            facility.focus()
            for polisher in polishers:
                polisher.move_to(facility)
            facility.polish_blueprint(names, polishers)

        for polisher in polishers:
            polisher.reap()

    def stop_all_blueprints(self, facilities=None):
        """
        Stops all nodes of the fittings plan

        :param facilities: explicit list of target facilities
        :type facilities: ``str`` or ``list`` of ``str``

        This function checks facilities to stop all nodes there.
        The default behaviour is to consider all facilities mentioned in the
        fittings plan. If a list of facilities is provided, than the action is
        limited to this subset only.

        This function has no effect on nodes that are already stopped.

        """

        plogging.info("Stopping nodes from all blueprints")

        if facilities is not None:
            facilities = self.list_facility(facilities)
        else:
            facilities = self.facilities

        for facility in facilities:
            facility.focus()
            facility.stop_all_blueprints()

    def stop_blueprint(self, names, facilities=None):
        """
        Stops nodes of one blueprint of the fittings plan

        :param names: the name(s) of the blueprint(s) to stop
        :type names: ``str`` or ``list`` of ``str``

        :param facilities: explicit list of target facilities
        :type facilities: ``str`` or ``list`` of ``str``

        This function checks facilities to stop nodes from some blueprint.
        The default behaviour is to consider all facilities mentioned in the
        fittings plan. If a list of facilities is provided, than the action is
        limited to this subset only.

        This function has no effect on nodes that are already stopped.

        """

        if isinstance(names, list):
            label = ' '.join(names)
        else:
            label = names

        plogging.info("Stopping nodes from blueprint '{}'".format(label))

        if facilities is not None:
            facilities = self.list_facility(facilities)
        else:
            facilities = self.facilities

        for facility in facilities:
            facility.focus()
            facility.stop_blueprint(names)

    def wipe_all_blueprints(self, facilities=None):
        """
        Destroys all nodes from fittings plan

        :param facilities: explicit list of target facilities
        :type facilities: ``str`` or ``list`` of ``str``

        This function checks facilities to destroy all nodes there.
        The default behaviour is to consider all facilities mentioned in the
        fittings plan. If a list of facilities is provided, than the action is
        limited to this subset only.

        Note:
            Running nodes are always preserved from destruction.
            Therefore the need to stop nodes, in a separate command, before
            they can be actually destroyed.

        """

        plogging.info("Wiping all blueprints")

        if facilities is not None:
            facilities = self.list_facility(facilities)
        else:
            facilities = self.facilities

        for facility in facilities:
            facility.focus()
            facility.wipe_all_blueprints()

    def wipe_blueprint(self, names, facilities=None):
        """
        Destroys nodes for one or several blueprint(s) of the fittings plan

        :param names: the name(s) of the blueprint(s) to destroy
        :type names: ``str`` or ``list`` of ``str``

        :param facilities: explicit list of target facilities
        :type facilities: ``str`` or ``list`` of ``str``

        This function checks facilities to destroy nodes from one blueprint.
        The default behaviour is to consider all facilities mentioned in the
        fittings plan. If a list of facilities is provided, than the action is
        limited to this subset only.

        Note:
            Running nodes are always preserved from destruction.
            Therefore the need to stop nodes, in a separate command, before
            they can be actually destroyed.

        """

        if isinstance(names, list):
            label = ' '.join(names)
        else:
            label = names

        plogging.info("Wiping blueprint '{}'".format(label))

        if facilities is not None:
            facilities = self.list_facility(facilities)
        else:
            facilities = self.facilities

        for facility in facilities:
            facility.focus()
            facility.wipe_blueprint(names)

    def destroy_all_blueprints(self, facilities=None):
        """
        Destroys all blueprints from fittings plan

        :param facilities: explicit list of target facilities
        :type facilities: ``str`` or ``list`` of ``str``

        This function checks facilities to destroy all blueprints there.
        The default behaviour is to consider all facilities mentioned in the
        fittings plan. If a list of facilities is provided, than the action is
        limited to this subset only.

        Note:
            Running nodes are always preserved from destruction.
            Therefore the need to stop nodes, in a separate command, before
            they can be actually destroyed.

        """

        plogging.info("Destroying all blueprints")

        if facilities is not None:
            facilities = self.list_facility(facilities)
        else:
            facilities = self.facilities

        for facility in facilities:
            facility.focus()
            facility.destroy_all_blueprints()

    def destroy_blueprint(self, names, facilities=None):
        """
        Destroys one or several blueprint(s) from fittings plan

        :param names: the name(s) of the blueprint(s) to destroy
        :type names: ``str`` or ``list`` of ``str``

        :param facilities: explicit list of target facilities
        :type facilities: ``str`` or ``list`` of ``str``

        This function checks facilities to destroy one single blueprint.
        The default behaviour is to consider all facilities mentioned in the
        fittings plan. If a list of facilities is provided, than the action is
        limited to this subset only.

        Note:
            Running nodes are always preserved from destruction.
            Therefore the need to stop nodes, in a separate command, before
            they can be actually destroyed.

        """

        if isinstance(names, list):
            label = ' '.join(names)
        else:
            label = names

        plogging.info("Destroying blueprint '{}'".format(label))

        if facilities is not None:
            facilities = self.list_facility(facilities)
        else:
            facilities = self.facilities

        for facility in facilities:
            facility.focus()
            facility.destroy_blueprint(names)

    def get_compute_driver(self, region=None, host=None):
        """
        Loads a compute driver from Apache Libcloud

        """

        driver = get_compute_driver(ComputeProvider.DIMENSIONDATA)

        instance = driver(
            key=self.get_user_name(),
            secret=self.get_user_password(),
            region=region,
            host=host)

        return instance

    def get_balancer_driver(self, region=None, host=None):
        """
        Loads a load balancer driver from Apache Libcloud

        """

        driver = get_balancer_driver(BalancerProvider.DIMENSIONDATA)

        instance = driver(
            key=self.get_user_name(),
            secret=self.get_user_password(),
            region=region,
            host=host)
        return instance

    def get_backup_driver(self, region=None, host=None):
        """
        Loads a backup driver from Apache Libcloud

        """

        driver = get_backup_driver(BackupProvider.DIMENSIONDATA)

        instance = driver(
            key=self.get_user_name(),
            secret=self.get_user_password(),
            region=region,
            host=host)
        return instance

    def lookup(self, token):
        """
        Retrieves the value attached to a token

        :param token: the token, e.g., 'sqlMaster.secret'
        :type token: ``str``

        :return: the value attached to this token, or `None`

        """

        if token == 'plumbery.version':
            return __version__

        if token == 'shared.user':
            return self.get_shared_user()

        if token == 'shared.secret':
            return self.get_shared_secret()

        if token in ('credentials.name', 'name.credentials'):
            return self.get_user_name()

        if token in ('credentials.password', 'password.credentials'):
            return self.get_user_password()

        if token.startswith('parameter.'):
            return self.get_parameter(token)

        if token.startswith('rsa_public.'):
            return self.get_rsa_secret(token)
        if token.endswith('.rsa_public'):
            return self.get_rsa_secret(token)
        if token.startswith('rsa_private.'):
            return self.get_rsa_secret(token)
        if token.endswith('.rsa_private'):
            return self.get_rsa_secret(token)

        if token.startswith('secret.'):
            return self.get_secret(token)
        if token.endswith('.secret'):
            return self.get_secret(token)

        if token.startswith('uuid.'):
            return self.get_secret(token)
        if token.endswith('.uuid'):
            return self.get_secret(token)

        if token.startswith('http://') or token.startswith('https://'):
            return self.get_secret(token)

        if token.startswith('environment.'):
            key = token[len('environment.'):]
            value = os.environ.get(key)
            if value is None:
                raise KeyError("'{}' is absent from environment".format(key))
            return value

        return None



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from plumbery.plogging import plogging

__all__ = ['PlumberyPolisher']


class PlumberyPolisher(object):
    """
    Polishes elements of the fittings plan

    :param settings: specific settings for this polisher
    :type param: ``dict``

    Even embedded fittings deserve some good treatment, and this what
    polishing is about. You know the usual sequence:

    * at some point, decide that all appliances have to be visited
    * find every appliance and prepare it
    * after the hard work, go back home for some rest

    This is exactly what plumbery is offering to you in a straightforward
    extensible mechanism.

        You can create polishers of your own, or use polishers from other
        persons. All polishers have to be placed in the directory
        ``plumbery.polishers``. Each polisher should be put in a separate
        python file, and define a class that repeats the file name. For example
        the file::

            plumbery\\polishers\\configure.py

        should contain::

            class ConfigurePolisher(PlumberyPolisher):
            ...

        Once this is done properly, you can use the polisher by mentioning it
        if the fittings plan used by plumbery. Also, you can pass any
        parameters that the polisher would require.

        Example of configuration of ``fittings.yaml``::

            ---
            safeMode: False
            actions:
              - configure:
                  file: nodes.yaml
              - ansible:
                  file: inventory.yaml
            ---
            # Frankfurt in Europe
            locationId: EU6
            regionId: dd-eu
            ...

        The engine provides multiple ways to polish nodes. For example::

            from plumbery.engine import PlumberyEngine
            PlumberyEngine('fittings.yaml').polish_all_nodes()

    """

    def __init__(self, settings):
        self.settings = settings

    @classmethod
    def filter(cls, polishers, filter=None):
        """
        Selects only the polisher you want, or take them all

        :param polishers: polishers to be applied
        :type polishers: list of :class:`plumbery.PlumberyPolisher`

        :param filter: the name of a single polisher to apply. If this
            parameter is missing, all polishers declared in the fittings plan
            will be applied
        :type filter: ``str``

        :return: a list of filtered polishers
        :rtype: ``list`` of :class:`plumbery.PlumberyPolisher` or ``[]``

        """
        if filter is None:
            for polisher in polishers:
                plogging.info("Using '{}'".format(
                    polisher.settings['name']))
            return polishers

        for polisher in polishers:
            if polisher.settings['name'] == filter:
                plogging.info("Using '{}'".format(filter))
                filtered = [polisher]
                return filtered

        # generate an exception if no implementation is available
        polisher = PlumberyPolisher.from_shelf(filter)

        plogging.info("Using '{}'".format(filter))
        return [polisher]

    @classmethod
    def from_shelf(cls, polishId, settings={}):
        """
        Picks up a polisher from the shelf

        :param polishId: name of the polisher to use, e.g., ``inventory``
        :type polishId: ``str``

        :param settings: specific settings for this polisher
        :type param: ``dict``

        :return: instance of a polisher ready to use
        :rtype: :class:`plumbery.PlumberyPolisher`

        """

        moduleName = 'plumbery.polishers.' + polishId
        polisherName = polishId.capitalize() + 'Polisher'
        try:

            plogging.debug("Importing '{}'".format(moduleName))
            polisherModule = __import__(
                moduleName,
                globals(),
                locals(),
                [polisherName])

            plogging.debug("Instantiating '{}'".format(polisherName))
            polisherClass = getattr(polisherModule, polisherName)
            if settings is None:
                settings = {}
            settings['name'] = polishId
            return polisherClass(settings)

        except ImportError as feedback:
            plogging.debug("Unable to find module '{}'".format(moduleName))
            raise feedback

        except Exception as feedback:
            plogging.debug("Unable to import '{}' from '{}'".format(
                polisherName, moduleName))
            raise feedback

    def go(self, engine):
        """
        Puts the shoes on, and go polishing

        :param engine: the plumbery engine itself
        :type engine: :class:`plumbery.PlumberyEngine`

        This function is called once, before starting the process of
        polishing each node. You can override it for any specific
        initialisation that you would require.

        """

        self.engine = engine

    def move_to(self, facility):
        """
        Enters a facility to polish nodes there

        :param facility: a target facility
        :type facility: :class:`plumbery.PlumberyFacility`

        This function is called once for each facility that is visited during
        the polishing process. You can override it for any specific
        initialisation that you would require.

        """

        self.facility = facility

    def shine_container(self, container):
        """
        prepares a container until it shines

        :param container: the container to be polished
        :type container: :class:`plumbery.PlumberyInfrastructure`

        This is where the hard work is done. You have to override this
        function in your own polisher. Note that you can compare the reality
        versus the theoritical settings if you want.

        """

        pass

    def shine_node(self, node, settings, container):
        """
        prepares a node until it shines

        :param node: the node to be polished
        :type node: :class:`libcloud.compute.base.Node`

        :param settings: the fittings plan for this node
        :type settings: ``dict``

        :param container: the container of this node
        :type container: :class:`plumbery.PlumberyInfrastructure`

        This is where the hard work is done. You have to override this
        function in your own polisher. Note that you can compare the reality
        versus the theoritical settings if you want.

        """

        pass

    def reap(self):
        """
        Reaps the outcome of all this polishing

        This function is called once, after all nodes have been polished. You
        can override it for any specific closure that you would require.

        """

        pass
